# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the converters module.
"""

import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import WCS
from numpy.testing import assert_allclose

from photutils.aperture import (CircularAnnulus, CircularAperture,
                                EllipticalAnnulus, EllipticalAperture,
                                RectangularAnnulus, RectangularAperture,
                                SkyCircularAnnulus, SkyCircularAperture,
                                SkyEllipticalAnnulus, SkyEllipticalAperture,
                                SkyRectangularAnnulus, SkyRectangularAperture)
from photutils.aperture.converters import (_scalar_aperture_to_region,
                                           _shapely_polygon_to_region,
                                           aperture_to_region,
                                           region_to_aperture)
from photutils.utils._optional_deps import HAS_REGIONS, HAS_SHAPELY


@pytest.fixture
def image_2d_wcs():
    return WCS(
        {
            'CTYPE1': 'RA---TAN',
            'CUNIT1': 'deg',
            'CDELT1': -0.0002777777778,
            'CRPIX1': 1,
            'CRVAL1': 337.5202808,
            'CTYPE2': 'DEC--TAN',
            'CUNIT2': 'deg',
            'CDELT2': 0.0002777777778,
            'CRPIX2': 1,
            'CRVAL2': -20.833333059999998,
        },
    )


def compare_region_shapes(reg1, reg2):
    from regions import PixCoord

    assert reg1.__class__ == reg2.__class__
    for param in reg1._params:
        par1 = getattr(reg1, param)
        par2 = getattr(reg2, param)
        if isinstance(par1, PixCoord):
            assert_allclose(par1.xy, par2.xy)
        elif isinstance(par1, SkyCoord):
            assert par1 == par2
        elif isinstance(par1, u.Quantity):
            assert_quantity_allclose(par1, par2)
        else:
            assert_allclose(par1, par2)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_translation_circle(image_2d_wcs):
    from regions import CirclePixelRegion, PixCoord

    region_shape = CirclePixelRegion(center=PixCoord(x=42, y=43), radius=4.2)
    aperture = region_to_aperture(region_shape)
    assert isinstance(aperture, CircularAperture)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.r, region_shape.radius)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = region_to_aperture(region_sky)
    assert isinstance(aperture_sky, SkyCircularAperture)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.r, region_sky.radius)

    # NOTE: If these no longer fail, we also have to account for
    # non-scalar inputs. Assume this is representative for the sky
    # counterpart too.
    match = 'must be a scalar PixCoord'
    with pytest.raises(ValueError, match=match):
        CirclePixelRegion(center=PixCoord(x=[0, 42], y=[1, 43]), radius=4.2)
    match = 'must be a strictly positive scalar'
    with pytest.raises(ValueError, match=match):
        CirclePixelRegion(center=PixCoord(x=42, y=43), radius=[1, 4.2])


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_translation_ellipse(image_2d_wcs):
    from regions import EllipsePixelRegion, PixCoord

    region_shape = EllipsePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10,
        angle=Angle(30, 'deg'),
    )
    aperture = region_to_aperture(region_shape)
    assert isinstance(aperture, EllipticalAperture)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.a * 2, region_shape.width)
    assert_allclose(aperture.b * 2, region_shape.height)
    assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = region_to_aperture(region_sky)
    assert isinstance(aperture_sky, SkyEllipticalAperture)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.a * 2, region_sky.width)
    assert_quantity_allclose(aperture_sky.b * 2, region_sky.height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg),
                             region_sky.angle)

    # NOTE: If these no longer fail, we also have to account for
    # non-scalar inputs. Assume this is representative for the sky
    # counterpart too.
    match = 'must be a scalar PixCoord'
    with pytest.raises(ValueError, match=match):
        EllipsePixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]),
            width=16,
            height=10,
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipsePixelRegion(
            center=PixCoord(x=42, y=43),
            width=[1, 16],
            height=10,
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipsePixelRegion(
            center=PixCoord(x=42, y=43),
            width=16,
            height=[1, 10],
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipsePixelRegion(
            center=PixCoord(x=42, y=43),
            width=16,
            height=10,
            angle=Angle([0, 30], 'deg'),
        )


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_translation_rectangle(image_2d_wcs):
    from regions import PixCoord, RectanglePixelRegion

    region_shape = RectanglePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10,
        angle=Angle(30, 'deg'),
    )
    aperture = region_to_aperture(region_shape)
    assert isinstance(aperture, RectangularAperture)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.w, region_shape.width)
    assert_allclose(aperture.h, region_shape.height)
    assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = region_to_aperture(region_sky)
    assert isinstance(aperture_sky, SkyRectangularAperture)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.w, region_sky.width)
    assert_quantity_allclose(aperture_sky.h, region_sky.height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg),
                             region_sky.angle)

    # NOTE: If these no longer fail, we also have to account for
    # non-scalar inputs. Assume this is representative for the sky
    # counterpart too.
    match = 'must be a scalar PixCoord'
    with pytest.raises(ValueError, match=match):
        RectanglePixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]),
            width=16,
            height=10,
            angle=Angle(30, 'deg'),
        )

    match = 'must be a strictly positive scalar'
    with pytest.raises(ValueError, match=match):
        RectanglePixelRegion(
            center=PixCoord(x=42, y=43),
            width=[1, 16],
            height=10,
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=match):
        RectanglePixelRegion(
            center=PixCoord(x=42, y=43),
            width=16,
            height=[1, 10],
            angle=Angle(30, 'deg'),
        )
    match = 'must be a scalar'
    with pytest.raises(ValueError, match=match):
        RectanglePixelRegion(
            center=PixCoord(x=42, y=43),
            width=16,
            height=10,
            angle=Angle([0, 30], 'deg'),
        )


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_translation_circle_annulus(image_2d_wcs):
    from regions import CircleAnnulusPixelRegion, PixCoord

    region_shape = CircleAnnulusPixelRegion(
        center=PixCoord(x=42, y=43), inner_radius=5, outer_radius=8,
    )
    aperture = region_to_aperture(region_shape)
    assert isinstance(aperture, CircularAnnulus)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.r_in, region_shape.inner_radius)
    assert_allclose(aperture.r_out, region_shape.outer_radius)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = region_to_aperture(region_sky)
    assert isinstance(aperture_sky, SkyCircularAnnulus)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.r_in, region_sky.inner_radius)
    assert_quantity_allclose(aperture_sky.r_out, region_sky.outer_radius)

    # NOTE: If these no longer fail, we also have to account for
    # non-scalar inputs. Assume this is representative for the sky
    # counterpart too.
    match = 'must be a scalar PixCoord'
    with pytest.raises(ValueError, match=match):
        CircleAnnulusPixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), inner_radius=5,
            outer_radius=8,
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        CircleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_radius=[1, 5], outer_radius=8,
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        CircleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_radius=5, outer_radius=[8, 10],
        )


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_translation_ellipse_annulus(image_2d_wcs):
    from regions import EllipseAnnulusPixelRegion, PixCoord

    region_shape = EllipseAnnulusPixelRegion(
        center=PixCoord(x=42, y=43),
        inner_width=5.5,
        inner_height=3.5,
        outer_width=8.5,
        outer_height=6.5,
        angle=Angle(30, 'deg'),
    )
    aperture = region_to_aperture(region_shape)
    assert isinstance(aperture, EllipticalAnnulus)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.a_in * 2, region_shape.inner_width)
    assert_allclose(aperture.a_out * 2, region_shape.outer_width)
    assert_allclose(aperture.b_in * 2, region_shape.inner_height)
    assert_allclose(aperture.b_out * 2, region_shape.outer_height)
    assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = region_to_aperture(region_sky)
    assert isinstance(aperture_sky, SkyEllipticalAnnulus)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.a_in * 2, region_sky.inner_width)
    assert_quantity_allclose(aperture_sky.a_out * 2, region_sky.outer_width)
    assert_quantity_allclose(aperture_sky.b_in * 2, region_sky.inner_height)
    assert_quantity_allclose(aperture_sky.b_out * 2, region_sky.outer_height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg),
                             region_sky.angle)

    # NOTE: If these no longer fail, we also have to account for
    # non-scalar inputs. Assume this is representative for the sky
    # counterpart too.
    match = 'must be a scalar PixCoord'
    with pytest.raises(ValueError, match=match):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]),
            inner_width=5.5,
            inner_height=3.5,
            outer_width=8.5,
            outer_height=6.5,
            angle=Angle(30, 'deg'),
        )
    match = 'must be a strictly positive scalar'
    with pytest.raises(ValueError, match=match):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=[1, 5.5],
            inner_height=3.5,
            outer_width=8.5,
            outer_height=6.5,
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=match):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=5.5,
            inner_height=[1, 3.5],
            outer_width=8.5,
            outer_height=6.5,
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=5.5,
            inner_height=3.5,
            outer_width=[8.5, 10],
            outer_height=6.5,
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=5.5,
            inner_height=3.5,
            outer_width=8.5,
            outer_height=[6.5, 10],
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=5.5,
            inner_height=3.5,
            outer_width=8.5,
            outer_height=6.5,
            angle=Angle([0, 30], 'deg'),
        )


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_translation_rectangle_annulus(image_2d_wcs):
    from regions import PixCoord, RectangleAnnulusPixelRegion

    region_shape = RectangleAnnulusPixelRegion(
        center=PixCoord(x=42, y=43),
        inner_width=5.5,
        inner_height=3.5,
        outer_width=8.5,
        outer_height=6.5,
        angle=Angle(30, 'deg'),
    )
    aperture = region_to_aperture(region_shape)
    assert isinstance(aperture, RectangularAnnulus)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.w_in, region_shape.inner_width)
    assert_allclose(aperture.w_out, region_shape.outer_width)
    assert_allclose(aperture.h_in, region_shape.inner_height)
    assert_allclose(aperture.h_out, region_shape.outer_height)
    assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = region_to_aperture(region_sky)
    assert isinstance(aperture_sky, SkyRectangularAnnulus)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.w_in, region_sky.inner_width)
    assert_quantity_allclose(aperture_sky.w_out, region_sky.outer_width)
    assert_quantity_allclose(aperture_sky.h_in, region_sky.inner_height)
    assert_quantity_allclose(aperture_sky.h_out, region_sky.outer_height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg),
                             region_sky.angle)

    # NOTE: If these no longer fail, we also have to account for
    # non-scalar inputs. Assume this is representative for the sky
    # counterpart too.
    match = 'must be a scalar PixCoord'
    with pytest.raises(ValueError, match=match):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]),
            inner_width=5.5,
            inner_height=3.5,
            outer_width=8.5,
            outer_height=6.5,
            angle=Angle(30, 'deg'),
        )

    match = 'must be a strictly positive scalar'
    with pytest.raises(ValueError, match=match):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=[1, 5.5],
            inner_height=3.5,
            outer_width=8.5,
            outer_height=6.5,
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=match):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=5.5,
            inner_height=[1, 3.5],
            outer_width=8.5,
            outer_height=6.5,
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=5.5,
            inner_height=3.5,
            outer_width=[8.5, 10],
            outer_height=6.5,
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=5.5,
            inner_height=3.5,
            outer_width=8.5,
            outer_height=[6.5, 10],
            angle=Angle(30, 'deg'),
        )
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43),
            inner_width=5.5,
            inner_height=3.5,
            outer_width=8.5,
            outer_height=6.5,
            angle=Angle([0, 30], 'deg'),
        )


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_translation_polygon():
    from regions import PixCoord, PolygonPixelRegion

    region_shape = PolygonPixelRegion(vertices=PixCoord(x=[1, 2, 2],
                                                        y=[1, 1, 2]))
    match = r'Cannot convert .* to an Aperture object'
    with pytest.raises(TypeError, match=match):
        region_to_aperture(region_shape)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_aperture_to_region():
    from regions import Region, Regions

    xypos = [(10, 20), (30, 40), (50, 60), (70, 80)]
    ra, dec = np.transpose(xypos)
    skycoord = SkyCoord(ra=ra, dec=dec, unit='deg')
    unit = u.arcsec

    apertures = [CircularAperture(xypos, r=3.0),
                 CircularAnnulus(xypos, r_in=3.0, r_out=7.0),
                 SkyCircularAperture(skycoord, r=3.0 * unit),
                 SkyCircularAnnulus(skycoord, r_in=3.0 * unit,
                                    r_out=7.0 * unit),
                 EllipticalAperture(xypos, a=10.0, b=5.0, theta=np.pi / 2.0),
                 EllipticalAnnulus(xypos, a_in=10.0, a_out=20.0, b_out=17.0,
                                   theta=np.pi / 3),
                 SkyEllipticalAperture(skycoord, a=10.0 * unit, b=5.0 * unit,
                                       theta=30 * u.deg),
                 SkyEllipticalAnnulus(skycoord, a_in=10.0 * unit,
                                      a_out=20.0 * unit, b_out=17.0 * unit,
                                      theta=60 * u.deg),
                 RectangularAperture(xypos, w=10.0, h=5.0, theta=np.pi / 2.0),
                 RectangularAnnulus(xypos, w_in=10.0, w_out=20.0, h_out=17,
                                    theta=np.pi / 3),
                 SkyRectangularAperture(skycoord, w=10.0 * unit, h=5.0 * unit,
                                        theta=30 * u.deg),
                 SkyRectangularAnnulus(skycoord, w_in=10.0 * unit,
                                       w_out=20.0 * unit, h_out=17.0 * unit,
                                       theta=60 * u.deg)]

    for aperture in apertures:
        region0 = aperture_to_region(aperture[0])
        region = aperture_to_region(aperture)

        assert isinstance(region0, Region)
        assert isinstance(region, Regions)
        assert len(region) == len(aperture)

        aper0 = region_to_aperture(region0)
        assert aper0 == aperture[0]


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_invalid_inputs():
    from regions import CirclePixelRegion, PixCoord

    aperture = CircularAperture((10, 12), r=4.2)
    region = CirclePixelRegion(center=PixCoord(x=10, y=12), radius=4.2)

    match = 'Input region must be a Region object'
    with pytest.raises(TypeError, match=match):
        region_to_aperture(aperture)

    match = 'Input aperture must be an Aperture object'
    with pytest.raises(TypeError, match=match):
        aperture_to_region(region)

    aperture = CircularAperture(((10, 12), (21, 7)), r=4.2)
    match = r'Only scalar .* apertures are supported'
    with pytest.raises(ValueError, match=match):
        _scalar_aperture_to_region(aperture)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
@pytest.mark.skipif(not HAS_SHAPELY, reason='shapely is required')
def test_shapely_polygon_to_region():
    from regions import PixCoord, PolygonPixelRegion
    from shapely import Polygon

    ref_region = PolygonPixelRegion(vertices=PixCoord(x=[1, 3, 2, 1],
                                                      y=[1, 1, 4, 2]))

    polygon = Polygon([(1, 1), (3, 1), (2, 4), (1, 2)])
    region = _shapely_polygon_to_region(polygon)
    assert region == ref_region

    match = 'Input must be a Polygon or MultiPolygon object'
    with pytest.raises(TypeError, match=match):
        _shapely_polygon_to_region('foo')


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
@pytest.mark.skipif(not HAS_SHAPELY, reason='shapely is required')
def test_shapely_multipolygon_to_region():
    """
    Test that _shapely_polygon_to_region handles MultiPolygon inputs by
    returning a Regions object containing one PolygonPixelRegion per
    polygon.
    """
    from regions import Regions
    from shapely import MultiPolygon, Polygon

    poly1 = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
    poly2 = Polygon([(2, 2), (3, 2), (3, 3), (2, 3)])
    multi = MultiPolygon([poly1, poly2])
    result = _shapely_polygon_to_region(multi)
    assert isinstance(result, Regions)
    assert len(result) == 2


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_scalar_aperture_to_region_unknown_type():
    """
    Test that _scalar_aperture_to_region raises TypeError for an
    aperture type that is not one of the 12 supported classes.
    """

    class _FakeAperture:
        """
        Minimal fake aperture with shape=() that passes the scalar check
        but has no matching isinstance branch.
        """

        shape = ()

    match = 'Cannot convert input aperture to a Region object'
    with pytest.raises(TypeError, match=match):
        _scalar_aperture_to_region(_FakeAperture())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Private utility functions for centroiding.
"""

import warnings

import numpy as np
from astropy.utils.exceptions import AstropyUserWarning


def _validate_data(data, *, ndim=2):
    """
    Validate and convert the input data array.

    Parameters
    ----------
    data : array_like
        The input data array.

    ndim : int or `None`, optional
        The required number of dimensions. If `None`, no dimensionality
        check is performed.

    Returns
    -------
    data : `~numpy.ndarray`
        The input data converted to a float array.
    """
    data = np.asanyarray(data, dtype=float)
    if ndim is not None and data.ndim != ndim:
        msg = f'data must be a {ndim}D array'
        raise ValueError(msg)
    return data


def _validate_mask_shape(data, mask):
    """
    Validate that the data and mask have the same shape.

    Parameters
    ----------
    data : `~numpy.ndarray`
        The input data array.

    mask : bool `~numpy.ndarray` or `None`
        The input mask array.
    """
    if mask is not None and data.shape != mask.shape:
        msg = 'data and mask must have the same shape'
        raise ValueError(msg)


def _process_data_mask(data, mask, *, ndim=2, fill_value=np.nan):
    """
    Process the input data and mask.

    This function validates the input data and mask, handles non-finite
    values, and returns the processed data. The input ``mask`` is never
    mutated; a new array is created if the mask needs to be combined
    with a `~numpy.ma.MaskedArray` mask. Copies of ``data`` are made
    only when modifications are required.

    Parameters
    ----------
    data : array_like
        The input data array.

    mask : bool `~numpy.ndarray` or `None`
        A boolean mask where `True` indicates a masked (invalid) pixel.

    ndim : int or `None`, optional
        The required number of dimensions for ``data``. If `None`, no
        dimensionality check is performed.

    fill_value : float, optional
        The value used to replace masked or non-finite data values.

    Returns
    -------
    data : `~numpy.ndarray`
        Processed data with masked and non-finite values replaced by
        ``fill_value``. Always returned as a plain `~numpy.ndarray`
        (never a `~numpy.ma.MaskedArray`).
    """
    data = _validate_data(data, ndim=ndim)
    is_copied = False
    is_masked_array = isinstance(data, np.ma.MaskedArray)
    _validate_mask_shape(data, mask)

    badmask = ~np.isfinite(data)

    if np.ma.is_masked(data):
        mask2 = data.mask
        mask = mask2 if mask is None else mask | mask2

    if mask is not None:
        if np.any(mask):
            data = data.copy()
            is_copied = True
            data[mask] = fill_value
        badmask &= ~mask

    if np.any(badmask):
        msg = ('Input data contains non-finite values (e.g., NaN or inf) '
               'that were automatically masked.')
        warnings.warn(msg, AstropyUserWarning)
        if not is_copied:
            data = data.copy()
        data[badmask] = fill_value

    # If the input was a MaskedArray, return a plain ndarray; the mask
    # has already been applied to the data above.
    if is_masked_array:
        data = np.asarray(data)

    return data


def _validate_gaussian_inputs(data, mask, error):
    """
    Process and validate the data, mask, and optional error inputs for
    Gaussian centroid functions.

    The input ``mask`` and ``error`` arrays are not mutated; copies are
    made only when modifications are needed.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The input data array.

    mask : 2D bool `~numpy.ndarray` or `None`
        A boolean mask where `True` indicates a masked (invalid) pixel.

    error : 2D `~numpy.ndarray` or `None`
        The 1-sigma error array.

    Returns
    -------
    data : 2D `~numpy.ndarray`
        Processed data with all invalid pixels (from the mask,
        non-finite data, and non-finite error) set to zero.

    combined_mask : 2D bool `~numpy.ndarray`
        Boolean mask of all invalid pixels.

    error : 2D `~numpy.ndarray` or `None`
        Error array with all invalid pixels set to zero (copied only
        if a modification was required), or `None` if no error was
        provided.
    """
    data = _process_data_mask(data, mask, fill_value=np.nan)
    combined_mask = ~np.isfinite(data)

    if error is not None:
        error = np.asanyarray(error, dtype=float)
        if data.shape != error.shape:
            msg = 'data and error must have the same shape'
            raise ValueError(msg)

        error_mask = ~np.isfinite(error)
        if np.any(error_mask):
            combined_mask |= error_mask

        # Zero error at all invalid pixel positions; copy only if needed
        if np.any(combined_mask):
            error = error.copy()
            error[combined_mask] = 0.0

    # Apply the full combined mask to data once
    if np.any(combined_mask):
        # The data array may still be the original input object if
        # no modifications were needed above. We copy here to avoid
        # mutating the caller's original data.
        data = data.copy()
        data[combined_mask] = 0.0

    return data, combined_mask, error


def _gaussian1d_moments(data, *, mask=None):
    """
    Estimate 1D Gaussian parameters from the moments of 1D data.

    This function can be useful for providing initial parameter values
    when fitting a 1D Gaussian to the ``data``.

    Parameters
    ----------
    data : 1D `~numpy.ndarray`
        The 1D data array.

    mask : 1D bool `~numpy.ndarray`, optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.

    Returns
    -------
    amplitude, mean, stddev : float
        The estimated parameters of a 1D Gaussian.
    """
    data = _process_data_mask(data, mask, ndim=1, fill_value=0.0)

    x = np.arange(data.size)
    x_mean = np.sum(x * data) / np.sum(data)
    x_stddev = np.sqrt(abs(np.sum(data * (x - x_mean) ** 2) / np.sum(data)))
    amplitude = np.ptp(data)

    return amplitude, x_mean, x_stddev


def _gaussian2d_moments(data):
    """
    Estimate 2D Gaussian parameters from the moments of 2D data.

    This function can be useful for providing initial parameter
    values when fitting a 2D Gaussian to the ``data``.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D data array. Values at masked pixels must already be set
        to zero before calling this function.

    Returns
    -------
    amplitude, x_mean, y_mean, x_stddev, y_stddev, theta : float
        The estimated amplitude (``amplitude``), centroid (``x_mean``,
        ``y_mean``), axis standard deviations (``x_stddev``,
        ``y_stddev``), and rotation angle in radians (``theta``) of a 2D
        Gaussian.
    """
    y, x = np.indices(data.shape)
    amplitude = np.max(data)
    total = np.sum(data)

    # 1st-order moments (centroid)
    x_mean = np.sum(x * data) / total
    y_mean = np.sum(y * data) / total

    # 2nd-order central moments
    dx = x - x_mean
    dy = y - y_mean
    mu_20 = np.sum(dx**2 * data) / total
    mu_02 = np.sum(dy**2 * data) / total
    mu_11 = np.sum(dx * dy * data) / total

    # Covariance matrix
    covar = np.array([[mu_02, mu_11], [mu_11, mu_20]])

    # Eigenvalues in descending order give semimajor/semiminor sigma^2
    eigvals = np.linalg.eigvalsh(covar)[::-1]
    eigvals = np.clip(eigvals, 0, None)
    x_stddev = np.sqrt(eigvals[0])
    y_stddev = np.sqrt(eigvals[1])

    # Orientation angle (radians) between x-axis and the major axis.
    # When the distribution is nearly isotropic (mu_20 ~ mu_02, mu_11 ~
    # 0), the angle is undefined; guard against floating-point noise by
    # returning theta = 0 in that case.
    anisotropy = np.sqrt((mu_20 - mu_02) ** 2 + (2.0 * mu_11) ** 2)
    if anisotropy < 1e-6 * (mu_20 + mu_02):
        theta = 0.0
    else:
        theta = 0.5 * np.arctan2(2.0 * mu_11, mu_20 - mu_02)

    return amplitude, x_mean, y_mean, x_stddev, y_stddev, theta



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Common test fixtures for the profiles module tests.
"""

import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D


@pytest.fixture(name='profile_data')
def fixture_profile_data():
    """
    Fixture that generates a 2D Gaussian profile with error and mask
    arrays for testing the profile classes.
    """
    xsize = 101
    ysize = 80
    xcen = (xsize - 1) / 2
    ycen = (ysize - 1) / 2
    xycen = (xcen, ycen)

    sig = 10.0
    model = Gaussian2D(21., xcen, ycen, sig, sig)
    y, x = np.mgrid[0:ysize, 0:xsize]
    data = model(x, y)

    error = 10.0 * np.sqrt(data)
    mask = np.zeros(data.shape, dtype=bool)
    mask[:int(ycen), :int(xcen)] = True

    return xycen, data, error, mask



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the core module.
"""

import astropy.units as u
import numpy as np
import pytest
from numpy.testing import assert_allclose

from photutils.datasets import make_wcs
from photutils.morphology import data_properties


def test_data_properties():
    """
    Test basics of ``data_properties`` with and without a mask.
    """
    data = np.ones((2, 2)).astype(float)
    mask = np.array([[False, False], [True, True]])
    props = data_properties(data, mask=None)
    props2 = data_properties(data, mask=mask)
    properties = ['x_centroid', 'y_centroid']
    result = [getattr(props, i) for i in properties]
    result2 = [getattr(props2, i) for i in properties]
    assert_allclose([0.5, 0.5], result, rtol=0, atol=1.0e-6)
    assert_allclose([0.5, 0.0], result2, rtol=0, atol=1.0e-6)
    assert props.area.value == 4.0
    assert props2.area.value == 2.0

    wcs = make_wcs(data.shape)
    props = data_properties(data, mask=None, wcs=wcs)
    assert props.sky_centroid is not None


def test_data_properties_invalid_data_shape():
    """
    Test that data must be a 2D array.
    """
    match = 'data must be a 2D array'
    with pytest.raises(ValueError, match=match):
        data_properties(np.ones(10))  # 1D

    with pytest.raises(ValueError, match=match):
        data_properties([1, 2, 3])  # 1D list

    with pytest.raises(ValueError, match=match):
        data_properties(np.ones((3, 3, 3)))  # 3D


def test_data_properties_mask_invalid_shape():
    """
    Test that mask must have the same shape as data.
    """
    data = np.ones((10, 10))

    match = 'mask must have the same shape as data'
    with pytest.raises(ValueError, match=match):
        data_properties(data, mask=np.zeros((5, 5), dtype=bool))

    match = 'mask must have the same shape as data'
    with pytest.raises(ValueError, match=match):
        data_properties(data, mask=np.zeros((10, 5), dtype=bool))


def test_data_properties_bkg():
    """
    Test with a scalar and 2D array background.
    """
    data = np.ones((3, 3)).astype(float)
    props = data_properties(data, background=1.0)
    assert props.area.value == 9.0
    assert props.background_sum == 9.0

    bkg_2d = np.full((3, 3), 2.0)
    props2 = data_properties(data, background=bkg_2d)
    assert props2.background_sum == 18.0


def test_data_properties_bkg_invalid():
    """
    Test that invalid background inputs raise ``ValueError``.
    """
    data = np.ones((3, 3))

    match = 'background must be a scalar or a 2D array'
    with pytest.raises(ValueError, match=match):
        data_properties(data, background=[1.0, 2.0])

    with pytest.raises(ValueError, match=match):
        data_properties(data, background=np.ones((2, 2)))


def test_data_properties_all_masked():
    """
    Test that an all-True mask raises ``ValueError``.
    """
    data = np.ones((4, 4))
    mask = np.ones((4, 4), dtype=bool)
    match = 'All pixels in data are masked'
    with pytest.raises(ValueError, match=match):
        data_properties(data, mask=mask)


def test_data_properties_quantity():
    """
    Test that ``~astropy.units.Quantity`` input is accepted.
    """
    data = np.ones((3, 3)) * u.Jy
    props = data_properties(data)
    assert props.area.value == 9.0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the model_helpers module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.fitting import TRFLSQFitter
from astropy.modeling.models import Const2D, Gaussian2D, Moffat2D
from astropy.nddata import NDData
from astropy.table import Table
from astropy.utils.exceptions import AstropyDeprecationWarning
from numpy.testing import assert_allclose, assert_equal

from photutils import datasets
from photutils.detection import find_peaks
from photutils.psf import (EPSFBuilder, extract_stars, grid_from_epsfs,
                           make_psf_model)
from photutils.psf.model_helpers import _integrate_model, _InverseShift


def test_inverse_shift():
    model = _InverseShift(10)
    assert model(1) == -9.0
    assert model(-10) == -20.0
    assert model.fit_deriv(10, 1)[0] == -1.0


def test_integrate_model():
    model = Gaussian2D(1, 5, 5, 1, 1) * Const2D(0.0)
    integral = _integrate_model(model, x_name='x_mean_0', y_name='y_mean_0')
    assert integral == 0.0

    integral = _integrate_model(model, x_name='x_mean_0', y_name='y_mean_0',
                                use_dblquad=True)
    assert integral == 0.0

    match = 'dx and dy must be > 0'
    with pytest.raises(ValueError, match=match):
        _integrate_model(model, x_name='x_mean_0', y_name='y_mean_0',
                         dx=-10, dy=10)
    with pytest.raises(ValueError, match=match):
        _integrate_model(model, x_name='x_mean_0', y_name='y_mean_0',
                         dx=10, dy=-10)

    match = 'subsample must be >= 1'
    with pytest.raises(ValueError, match=match):
        _integrate_model(model, x_name='x_mean_0', y_name='y_mean_0',
                         subsample=-1)

    match = 'model x and y positions must be finite'
    model = Gaussian2D(1, np.inf, 5, 1, 1)
    with pytest.raises(ValueError, match=match):
        _integrate_model(model, x_name='x_mean', y_name='y_mean')


@pytest.fixture(name='moffat_source', scope='module')
def fixture_moffat_source():
    model = Moffat2D(alpha=4.8)

    # this is the analytic value needed to get a total flux of 1
    model.amplitude = (model.alpha - 1.0) / (np.pi * model.gamma**2)

    xx, yy = np.meshgrid(*([np.linspace(-2, 2, 100)] * 2))

    return model, (xx, yy, model(xx, yy))


def test_moffat_fitting(moffat_source):
    """
    Test fitting with a Moffat2D model.
    """
    model, (xx, yy, data) = moffat_source

    # initial Moffat2D model close to the original
    guess_moffat = Moffat2D(x_0=0.1, y_0=-0.05, gamma=1.05,
                            amplitude=model.amplitude * 1.06, alpha=4.75)

    fitter = TRFLSQFitter()
    fit = fitter(guess_moffat, xx, yy, data)
    assert_allclose(fit.parameters, model.parameters, rtol=0.01, atol=0.0005)


# we set the tolerances in flux to be 2-3% because the guessed model
# parameters are known to be wrong
@pytest.mark.parametrize(('kwargs', 'tols'),
                         [({'x_name': 'x_0', 'y_name': 'y_0',
                            'flux_name': None, 'normalize': True},
                           (1e-3, 0.02)),
                          ({'x_name': None, 'y_name': None, 'flux_name': None,
                            'normalize': True}, (1e-3, 0.02)),
                          ({'x_name': None, 'y_name': None, 'flux_name': None,
                            'normalize': False}, (1e-3, 0.03)),
                          ({'x_name': 'x_0', 'y_name': 'y_0',
                            'flux_name': 'amplitude', 'normalize': False},
                           (1e-3, None))])
def test_make_psf_model(moffat_source, kwargs, tols):
    model, (xx, yy, data) = moffat_source

    # a close-but-wrong "guessed Moffat"
    guess_moffat = Moffat2D(x_0=0.1, y_0=-0.05, gamma=1.01,
                            amplitude=model.amplitude * 1.01, alpha=4.79)
    if kwargs['normalize']:
        # definitely very wrong, so this ensures the renormalization
        # works
        guess_moffat.amplitude = 5.0

    if kwargs['x_name'] is None:
        guess_moffat.x_0 = 0
    if kwargs['y_name'] is None:
        guess_moffat.y_0 = 0

    psf_model = make_psf_model(guess_moffat, **kwargs)
    fitter = TRFLSQFitter()
    fit_model = fitter(psf_model, xx, yy, data)
    xytol, fluxtol = tols

    if xytol is not None:
        assert np.abs(getattr(fit_model, fit_model.x_name)) < xytol
        assert np.abs(getattr(fit_model, fit_model.y_name)) < xytol
    if fluxtol is not None:
        assert np.abs(1.0 - getattr(fit_model, fit_model.flux_name)) < fluxtol

    # ensure the model parameters did not change
    assert fit_model[2].gamma == guess_moffat.gamma
    assert fit_model[2].alpha == guess_moffat.alpha
    if kwargs['flux_name'] is None:
        assert fit_model[2].amplitude == guess_moffat.amplitude


def test_make_psf_model_units():
    model = Moffat2D(amplitude=1.0 * u.Jy, x_0=25, y_0=25, alpha=4.8,
                     gamma=3.1)
    model.amplitude = (model.amplitude.unit * (model.alpha - 1.0)
                       / (np.pi * model.gamma**2))  # normalize to flux=1

    psf_model = make_psf_model(model, x_name='x_0', y_name='y_0',
                               normalize=True)
    yy, xx = np.mgrid[:51, :51]
    data1 = model(xx, yy)
    data2 = psf_model(xx, yy)
    assert_allclose(data1, data2)


def test_make_psf_model_compound():
    model = (Const2D(0.0) + Const2D(1.0) + Gaussian2D(1, 5, 5, 1, 1)
             * Const2D(1.0) * Const2D(1.0))
    psf_model = make_psf_model(model, x_name='x_mean_2', y_name='y_mean_2',
                               normalize=True)
    assert psf_model.x_name == 'x_mean_4'
    assert psf_model.y_name == 'y_mean_4'
    assert psf_model.flux_name == 'amplitude_7'


def test_make_psf_model_inputs():
    model = Gaussian2D(1, 5, 5, 1, 1)
    match = 'parameter name not found in the input model'
    with pytest.raises(ValueError, match=match):
        make_psf_model(model, x_name='x_mean_0', y_name='y_mean')
    with pytest.raises(ValueError, match=match):
        make_psf_model(model, x_name='x_mean', y_name='y_mean_10')


def test_make_psf_model_integral():
    model = Gaussian2D(1, 5, 5, 1, 1) * Const2D(0.0)
    match = 'Cannot normalize the model because the integrated flux is zero'
    with pytest.raises(ValueError, match=match):
        make_psf_model(model, x_name='x_mean_0', y_name='y_mean_0',
                       normalize=True)


def test_make_psf_model_offset():
    """
    Test to ensure the offset is in the correct direction.
    """
    moffat = Moffat2D(x_0=0, y_0=0, alpha=4.8)
    psfmod1 = make_psf_model(moffat.copy(), x_name='x_0', y_name='y_0',
                             normalize=False)
    psfmod2 = make_psf_model(moffat.copy(), normalize=False)
    moffat.x_0 = 10
    psfmod1.x_0_2 = 10
    psfmod2.offset_0 = 10

    assert moffat(10, 0) == psfmod1(10, 0) == psfmod2(10, 0) == 1.0


@pytest.mark.remote_data
class TestGridFromEPSFs:
    """
    Tests for `photutils.psf.utils.grid_from_epsfs`.
    """

    def setup_class(self, *, cutout_size=25):
        # make a set of 4 EPSF models

        self.cutout_size = cutout_size

        # make simulated image
        hdu = datasets.load_simulated_hst_star_image()
        data = hdu.data

        # break up the image into four quadrants
        q1 = data[0:500, 0:500]
        q2 = data[0:500, 500:1000]
        q3 = data[500:1000, 0:500]
        q4 = data[500:1000, 500:1000]

        # select some starts from each quadrant to use to build the epsf
        quad_stars = {'q1': {'data': q1, 'fiducial': (0., 0.), 'epsf': None},
                      'q2': {'data': q2, 'fiducial': (1000., 1000.),
                             'epsf': None},
                      'q3': {'data': q3, 'fiducial': (1000., 0.),
                             'epsf': None},
                      'q4': {'data': q4, 'fiducial': (0., 1000.),
                             'epsf': None}}

        for q in ['q1', 'q2', 'q3', 'q4']:
            quad_data = quad_stars[q]['data']
            peaks_tbl = find_peaks(quad_data, threshold=500.)

            # filter out sources near edge
            size = cutout_size
            hsize = (size - 1) / 2
            x = peaks_tbl['x_peak']
            y = peaks_tbl['y_peak']
            mask = ((x > hsize) & (x < (quad_data.shape[1] - 1 - hsize))
                    & (y > hsize) & (y < (quad_data.shape[0] - 1 - hsize)))

            stars_tbl = Table()
            stars_tbl['x'] = peaks_tbl['x_peak'][mask]
            stars_tbl['y'] = peaks_tbl['y_peak'][mask]

            stars = extract_stars(NDData(quad_data), stars_tbl,
                                  size=cutout_size)

            epsf_builder = EPSFBuilder(oversampling=4, maxiters=3,
                                       progress_bar=False)
            epsf, _ = epsf_builder(stars)

            # set x_0, y_0 to fiducial point
            epsf.y_0 = quad_stars[q]['fiducial'][0]
            epsf.x_0 = quad_stars[q]['fiducial'][1]

            quad_stars[q]['epsf'] = epsf

        self.epsfs = [val['epsf'] for val in quad_stars.values()]
        self.grid_xypos = [val['fiducial'] for val in quad_stars.values()]

    def test_basic_test_grid_from_epsfs(self):
        with pytest.warns(AstropyDeprecationWarning):
            psf_grid = grid_from_epsfs(self.epsfs)

        assert np.all(psf_grid.oversampling == self.epsfs[0].oversampling)
        assert psf_grid.data.shape == (4, psf_grid.oversampling[0] * 25 + 1,
                                       psf_grid.oversampling[1] * 25 + 1)

    def test_grid_xypos(self):
        """
        Test both options for setting PSF locations.
        """
        # default option x_0 and y_0s on input EPSFs
        with pytest.warns(AstropyDeprecationWarning):
            psf_grid = grid_from_epsfs(self.epsfs)

        assert psf_grid.meta['grid_xypos'] == [(0.0, 0.0), (1000.0, 1000.0),
                                               (0.0, 1000.0), (1000.0, 0.0)]

        # or pass in a list
        grid_xypos = [(250.0, 250.0), (750.0, 750.0),
                      (250.0, 750.0), (750.0, 250.0)]

        with pytest.warns(AstropyDeprecationWarning):
            psf_grid = grid_from_epsfs(self.epsfs, grid_xypos=grid_xypos)
        assert psf_grid.meta['grid_xypos'] == grid_xypos

    def test_meta(self):
        """
        Test the option for setting 'meta'.
        """
        keys = ['grid_xypos', 'oversampling', 'fill_value']

        # when 'meta' isn't provided, there should be just three keys
        with pytest.warns(AstropyDeprecationWarning):
            psf_grid = grid_from_epsfs(self.epsfs)
        for key in keys:
            assert key in psf_grid.meta

        # when meta is provided, those new keys should exist and anything
        # in the list above should be overwritten
        meta = {'grid_xypos': 0.0, 'oversampling': 0.0,
                'fill_value': -999, 'extra_key': 'extra'}
        with pytest.warns(AstropyDeprecationWarning):
            psf_grid = grid_from_epsfs(self.epsfs, meta=meta)
        for key in [*keys, 'extra_key']:
            assert key in psf_grid.meta
        assert psf_grid.meta['grid_xypos'].sort() == self.grid_xypos.sort()
        assert_equal(psf_grid.meta['oversampling'], [4, 4])
        assert psf_grid.meta['fill_value'] == 0.0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the radial_profile module.
"""

import warnings

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.models import Gaussian1D, Moffat1D
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose, assert_equal

from photutils.aperture import CircularAnnulus, CircularAperture
from photutils.profiles import RadialProfile
from photutils.utils._optional_deps import HAS_MATPLOTLIB


class TestRadialProfile:
    def test_basic(self, profile_data):
        """
        Test basic RadialProfile functionality.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp1 = RadialProfile(data, xycen, edge_radii, error=None, mask=None)

        assert_equal(rp1.radius, np.arange(35) + 0.5)
        assert rp1.area.shape == (35,)
        assert rp1.profile.shape == (35,)
        assert rp1.profile_error.shape == (0,)
        assert rp1.area[0] > 0.0

        assert len(rp1.apertures) == 35
        assert isinstance(rp1.apertures[0], CircularAperture)
        assert isinstance(rp1.apertures[1], CircularAnnulus)

        edge_radii = np.arange(36) + 0.1
        rp2 = RadialProfile(data, xycen, edge_radii, error=None, mask=None)
        assert isinstance(rp2.apertures[0], CircularAnnulus)

    def test_normalization(self, profile_data):
        """
        Test RadialProfile normalize and unnormalize methods.
        """
        xycen, data, error, _ = profile_data

        edge_radii = np.arange(36)
        rp1 = RadialProfile(data, xycen, edge_radii, error=error, mask=None)

        profile = rp1.profile
        profile_error = rp1.profile_error
        data_profile = rp1.data_profile
        rp1.normalize()
        assert np.max(rp1.profile) == 1.0
        assert np.max(rp1.profile_error) <= np.max(profile_error)
        assert np.max(rp1.data_profile) <= np.max(data_profile)

        rp1.unnormalize()
        assert_allclose(rp1.profile, profile)
        assert_allclose(rp1.profile_error, profile_error)
        assert_allclose(rp1.data_profile, data_profile)

    def test_data(self, profile_data):
        """
        Test RadialProfile data_radius and data_profile attributes.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp1 = RadialProfile(data, xycen, edge_radii, error=None, mask=None)

        data_radius = rp1.data_radius
        data_profile = rp1.data_profile
        assert np.max(data_radius) <= np.max(edge_radii)
        assert data_radius.shape == data_profile.shape
        assert np.min(data_profile) >= np.min(data)
        assert np.max(data_profile) <= np.max(data)

    def test_data_mask(self, profile_data):
        """
        Test that masked pixels are excluded from data_radius and
        data_profile.
        """
        xycen, data, _, mask = profile_data

        edge_radii = np.arange(36)
        rp1 = RadialProfile(data, xycen, edge_radii, mask=None)
        rp2 = RadialProfile(data, xycen, edge_radii, mask=mask)

        # Applying a mask should reduce the number of data points returned
        assert len(rp2.data_radius) < len(rp1.data_radius)
        assert rp2.data_radius.shape == rp2.data_profile.shape

        # Auto-masked non-finite pixels should also be excluded
        xcen, ycen = xycen
        data2 = data.copy()
        data2[int(ycen), int(xcen)] = np.nan
        match = 'Input data contains non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            rp3 = RadialProfile(data2, xycen, edge_radii, mask=None)
        assert np.all(np.isfinite(rp3.data_profile))
        assert len(rp3.data_profile) < len(rp1.data_profile)

    def test_inputs(self, profile_data):
        """
        Test RadialProfile input validation.
        """
        xycen, data, _, _ = profile_data

        match = 'minimum radii must be >= 0'
        edge_radii = np.arange(-1, 10)
        with pytest.raises(ValueError, match=match):
            RadialProfile(data, xycen, edge_radii, error=None, mask=None)

        match = 'radii must be a 1D array and have at least two values'
        edge_radii = [1]
        with pytest.raises(ValueError, match=match):
            RadialProfile(data, xycen, edge_radii, error=None, mask=None)

        edge_radii = np.arange(6).reshape(2, 3)
        with pytest.raises(ValueError, match=match):
            RadialProfile(data, xycen, edge_radii, error=None, mask=None)

        match = 'radii must be strictly increasing'
        edge_radii = np.arange(10)[::-1]
        with pytest.raises(ValueError, match=match):
            RadialProfile(data, xycen, edge_radii, error=None, mask=None)

        match = 'error must have the same shape as data'
        edge_radii = np.arange(10)
        with pytest.raises(ValueError, match=match):
            RadialProfile(data, xycen, edge_radii, error=np.ones(3),
                          mask=None)

        match = 'mask must have the same shape as data'
        edge_radii = np.arange(10)
        mask = np.ones(3, dtype=bool)
        with pytest.raises(ValueError, match=match):
            RadialProfile(data, xycen, edge_radii, error=None, mask=mask)

    def test_gaussian(self, profile_data):
        """
        Test RadialProfile Gaussian fit attributes.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp1 = RadialProfile(data, xycen, edge_radii, error=None, mask=None)

        assert isinstance(rp1.gaussian_fit, Gaussian1D)
        assert rp1.gaussian_profile.shape == (35,)
        assert rp1.gaussian_fwhm < 23.6

        edge_radii = np.arange(201)
        rp2 = RadialProfile(data, xycen, edge_radii, error=None, mask=None)
        assert isinstance(rp2.gaussian_fit, Gaussian1D)
        assert rp2.gaussian_profile.shape == (200,)
        assert rp2.gaussian_fwhm < 23.6

    def test_unit(self, profile_data):
        """
        Test RadialProfile with units.
        """
        xycen, data, error, _ = profile_data

        edge_radii = np.arange(36)
        unit = u.Jy
        rp1 = RadialProfile(data << unit, xycen, edge_radii,
                            error=error << unit, mask=None)
        assert rp1.profile.unit == unit
        assert rp1.profile_error.unit == unit

        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            RadialProfile(data << unit, xycen, edge_radii, error=error,
                          mask=None)

    def test_no_mutation(self, profile_data):
        """
        Test that input data, error, mask, and radii arrays are not
        mutated by RadialProfile.
        """
        xycen, data, error, mask = profile_data

        # Introduce a NaN to trigger the badmask / mask-merge code path.
        # Use a pixel outside the fixture's masked region (mask[:39, :50]).
        data2 = data.copy()
        data2[50, 70] = np.nan
        mask2 = mask.copy()
        data2_orig = data2.copy()
        error_orig = error.copy()
        mask2_orig = mask2.copy()
        edge_radii = np.arange(36)
        radii_orig = edge_radii.copy()

        match = 'Input data contains non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            RadialProfile(data2, xycen, edge_radii, error=error, mask=mask2)

        assert_equal(data2, data2_orig)
        assert_equal(error, error_orig)
        assert_equal(mask2, mask2_orig)
        assert_equal(edge_radii, radii_orig)

    def test_error(self, profile_data):
        """
        Test RadialProfile with error array.
        """
        xycen, data, error, _ = profile_data

        edge_radii = np.arange(36)
        rp1 = RadialProfile(data, xycen, edge_radii, error=error, mask=None)

        assert_equal(rp1.radius, np.arange(35) + 0.5)
        assert rp1.area.shape == (35,)
        assert rp1.profile.shape == (35,)
        assert rp1.profile_error.shape == (35,)

        assert len(rp1.apertures) == 35
        assert isinstance(rp1.apertures[0], CircularAperture)
        assert isinstance(rp1.apertures[1], CircularAnnulus)

    def test_gaussian_zero_sum(self, profile_data):
        """
        Test that ``gaussian_fit`` issues a warning and falls back to
        ``std=1.0`` when the profile sum is zero (covers the
        ``sum_profile == 0`` branch in ``gaussian_fit``).
        """
        xycen, data, _, _ = profile_data

        # All-zero data produces a zero-sum profile
        zero_data = np.zeros_like(data)
        edge_radii = np.arange(36)
        rp = RadialProfile(zero_data, xycen, edge_radii)

        with pytest.warns(AstropyUserWarning) as warning_list:
            gfit = rp.gaussian_fit

        messages = [str(w.message) for w in warning_list]
        assert any('The profile sum is zero' in m for m in messages)

        # The fit should still return a Gaussian1D model
        assert isinstance(gfit, Gaussian1D)

    def test_normalize_nan(self, profile_data):
        """
        If the profile has NaNs (e.g., aperture outside the image), make
        sure the normalization ignores them.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(101)
        rp1 = RadialProfile(data, xycen, edge_radii)
        rp1.normalize()
        assert not np.isnan(rp1.profile[0])

    def test_nonfinite(self, profile_data):
        """
        Test RadialProfile handling of non-finite data values.
        """
        xycen, data, error, _ = profile_data
        data2 = data.copy()
        data2[40, 40] = np.nan
        mask = ~np.isfinite(data2)

        edge_radii = np.arange(36)
        rp1 = RadialProfile(data, xycen, edge_radii, error=None, mask=mask)

        rp2 = RadialProfile(data2, xycen, edge_radii, error=error, mask=mask)
        assert_allclose(rp1.profile, rp2.profile)

        match = 'Input data contains non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            rp3 = RadialProfile(data2, xycen, edge_radii, error=error,
                                mask=None)
        assert_allclose(rp1.profile, rp3.profile)

        error2 = error.copy()
        error2[40, 40] = np.inf
        with pytest.warns(AstropyUserWarning, match=match):
            rp4 = RadialProfile(data, xycen, edge_radii, error=error2,
                                mask=None)
        assert_allclose(rp1.profile, rp4.profile)

    def test_all_masked(self, profile_data):
        """
        Test RadialProfile with all data masked.

        When every pixel is masked the profile should be all NaN
        (division by zero area).
        """
        xycen, data, _, _ = profile_data

        all_mask = np.ones(data.shape, dtype=bool)
        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii, error=None, mask=all_mask)
        assert np.all(np.isnan(rp.profile))

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot(self, profile_data):
        """
        Test RadialProfile plot methods.
        """
        xycen, data, error, _ = profile_data

        edge_radii = np.arange(36)
        rp1 = RadialProfile(data, xycen, edge_radii, error=None)
        rp1.plot()
        match = 'Errors were not input'
        with pytest.warns(AstropyUserWarning, match=match):
            rp1.plot_error()

        rp2 = RadialProfile(data, xycen, edge_radii, error=error)
        rp2.plot()
        pc1 = rp2.plot_error()
        assert_allclose(pc1.get_facecolor(), [[0.5, 0.5, 0.5, 0.3]])
        pc2 = rp2.plot_error(facecolor='blue')
        assert_allclose(pc2.get_facecolor(), [[0, 0, 1, 1]])

        unit = u.Jy
        rp3 = RadialProfile(data << unit, xycen, edge_radii,
                            error=error << unit)
        rp3.plot()
        rp3.plot_error()

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot_error_none(self, profile_data):
        """
        Test that ``plot_error()`` returns ``None`` when no errors were
        input.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii, error=None)
        match = 'Errors were not input'
        with pytest.warns(AstropyUserWarning, match=match):
            result = rp.plot_error()
        assert result is None

    def test_gaussian_std_guard(self, profile_data):
        """
        Test that the ``std = max(std, 1.0)`` guard is exercised when
        the computed std from a near-delta-function profile is close to
        zero.
        """
        xycen, data, _, _ = profile_data

        # Create a near-delta-function: all zero except one pixel at the
        # center so the weighted mean radius is nearly zero.
        delta = np.zeros_like(data)
        cy, cx = round(xycen[1]), round(xycen[0])
        delta[cy, cx] = 1.0

        edge_radii = np.arange(36)
        rp = RadialProfile(delta, xycen, edge_radii)
        # The radii warning may also be emitted because the fitted
        # stddev can be small enough that radius.min() > 0.3 * stddev.
        with pytest.warns(AstropyUserWarning):
            gfit = rp.gaussian_fit
        # The initial std guess gets clamped to 1.0, and the fit should
        # still produce a valid Gaussian
        assert isinstance(gfit, Gaussian1D)
        assert gfit.stddev.value > 0

    def test_gaussian_radii_warning(self, profile_data):
        """
        Test that a warning is issued when the input radii do not extend
        close to the source center.
        """
        xycen, data, _, _ = profile_data

        # Use edge_radii that start far from the center so that
        # radius.min() > 0.3 * stddev.
        edge_radii = np.arange(20, 36)
        rp = RadialProfile(data, xycen, edge_radii)
        match = 'Gaussian fit may be unreliable'
        with pytest.warns(AstropyUserWarning, match=match):
            _ = rp.gaussian_fit

    def test_repr(self, profile_data):
        """
        Test __repr__ output format.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii)
        r = repr(rp)
        assert 'RadialProfile' in r
        assert f'xycen={xycen}' in r
        assert f'n_radii={len(edge_radii)}' in r
        assert 'normalized=False' in r

        rp.normalize()
        r = repr(rp)
        assert 'normalized=True' in r

    def test_gaussian_fit_all_masked(self, profile_data):
        """
        Test that gaussian_fit returns None when the profile is entirely
        masked.
        """
        xycen, data, _, _ = profile_data

        mask = np.ones(data.shape, dtype=bool)
        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii, mask=mask)
        match = 'The radial profile is entirely non-finite or masked'
        with pytest.warns(AstropyUserWarning, match=match):
            result = rp.gaussian_fit
        assert result is None
        assert rp.gaussian_profile is None
        assert rp.gaussian_fwhm is None

    def test_normalize_all_nan(self, profile_data):
        """
        Test that normalize warns when the profile is all NaN.
        """
        xycen, data, _, _ = profile_data

        mask = np.ones(data.shape, dtype=bool)
        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii, mask=mask)
        match = 'The profile cannot be normalized'
        with pytest.warns(AstropyUserWarning, match=match):
            rp.normalize()

    def test_moffat(self, profile_data):
        """
        Test RadialProfile Moffat fit attributes.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii, error=None, mask=None)

        assert isinstance(rp.moffat_fit, Moffat1D)
        assert rp.moffat_profile.shape == (35,)
        assert rp.moffat_fwhm > 0
        assert rp.moffat_fwhm < 30.0

        # Check that x_0 is fixed at 0
        assert rp.moffat_fit.x_0.value == 0.0

        edge_radii = np.arange(201)
        rp2 = RadialProfile(data, xycen, edge_radii, error=None, mask=None)
        assert isinstance(rp2.moffat_fit, Moffat1D)
        assert rp2.moffat_profile.shape == (200,)
        assert rp2.moffat_fwhm > 0

    def test_moffat_fwhm_consistency(self, profile_data):
        """
        Test that ``moffat_fwhm`` is consistent with
        ``moffat_fit.fwhm``.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii)
        assert_allclose(rp.moffat_fwhm, rp.moffat_fit.fwhm)

    def test_moffat_profile_values(self, profile_data):
        """
        Test that ``moffat_profile`` equals the fit model evaluated at
        the profile radii.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii)
        expected = rp.moffat_fit(rp.radius)
        assert_allclose(rp.moffat_profile, expected)

    def test_moffat_zero_sum(self, profile_data):
        """
        Test that ``moffat_fit`` issues a warning and falls back to
        ``gamma=1.0`` when the profile sum is zero.
        """
        xycen, data, _, _ = profile_data

        zero_data = np.zeros_like(data)
        edge_radii = np.arange(36)
        rp = RadialProfile(zero_data, xycen, edge_radii)

        with pytest.warns(AstropyUserWarning) as warning_list:
            mfit = rp.moffat_fit

        messages = [str(w.message) for w in warning_list]
        assert any('The profile sum is zero' in m for m in messages)
        assert isinstance(mfit, Moffat1D)

    def test_moffat_fit_all_masked(self, profile_data):
        """
        Test that ``moffat_fit`` returns ``None`` when the profile is
        entirely masked.
        """
        xycen, data, _, _ = profile_data

        mask = np.ones(data.shape, dtype=bool)
        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii, mask=mask)
        match = 'The radial profile is entirely non-finite or masked'
        with pytest.warns(AstropyUserWarning, match=match):
            result = rp.moffat_fit
        assert result is None
        assert rp.moffat_profile is None
        assert rp.moffat_fwhm is None

    def test_moffat_no_above_half_max(self, profile_data):
        """
        Test that ``moffat_fit`` handles the case where no profile
        values are above half the maximum (gamma fallback to 1.0).
        """
        xycen, data, _, _ = profile_data

        # Create a near-delta-function profile so that all annular
        # averages may be below half-max (only center pixel has flux).
        delta = np.zeros_like(data)
        cy, cx = round(xycen[1]), round(xycen[0])
        delta[cy, cx] = 1.0

        edge_radii = np.arange(36)
        rp = RadialProfile(delta, xycen, edge_radii)
        # The fit may warn about convergence for such a narrow profile.
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', AstropyUserWarning)
            mfit = rp.moffat_fit
        assert isinstance(mfit, Moffat1D)
        assert mfit.gamma.value > 0
        assert mfit.alpha.value >= 1

    def test_moffat_lazyproperty(self, profile_data):
        """
        Test that ``moffat_fit``, ``moffat_profile``, and
        ``moffat_fwhm`` are lazily computed and cached.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii)

        fit1 = rp.moffat_fit
        fit2 = rp.moffat_fit
        assert fit1 is fit2

        prof1 = rp.moffat_profile
        prof2 = rp.moffat_profile
        assert prof1 is prof2

        fwhm1 = rp.moffat_fwhm
        fwhm2 = rp.moffat_fwhm
        assert fwhm1 == fwhm2

    def test_moffat_normalized(self, profile_data):
        """
        Test that normalizing the profile invalidates the Moffat fit
        cache and the fit is recomputed on the normalized profile.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii)

        amp_before = rp.moffat_fit.amplitude.value
        rp.normalize()
        # After normalization, the fit should be recomputed on the
        # normalized profile with max ~1.0, so the amplitude should
        # differ from the unnormalized fit.
        amp_after = rp.moffat_fit.amplitude.value
        assert amp_after != pytest.approx(amp_before, rel=0.1)
        assert amp_after == pytest.approx(1.0, abs=0.2)

    def test_moffat_with_error(self, profile_data):
        """
        Test Moffat fit with error input.
        """
        xycen, data, error, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii, error=error)
        assert isinstance(rp.moffat_fit, Moffat1D)
        assert rp.moffat_profile.shape == rp.profile.shape
        assert rp.moffat_fwhm > 0

    def test_moffat_with_mask(self, profile_data):
        """
        Test Moffat fit with a partial mask.
        """
        xycen, data, _, mask = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii, mask=mask)
        assert isinstance(rp.moffat_fit, Moffat1D)
        assert rp.moffat_fwhm > 0

    def test_moffat_nonfinite_data(self, profile_data):
        """
        Test Moffat fit with non-finite data values.
        """
        xycen, data, error, _ = profile_data

        data2 = data.copy()
        data2[40, 40] = np.nan
        edge_radii = np.arange(36)
        match = 'Input data contains non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            rp = RadialProfile(data2, xycen, edge_radii, error=error)

        assert isinstance(rp.moffat_fit, Moffat1D)
        assert rp.moffat_fwhm > 0

    def test_gaussian_fit_invalidated_on_normalize(self, profile_data):
        """
        Test that Gaussian fit properties are invalidated when the
        profile is normalized, and the recomputed fit matches the
        normalized profile.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii)

        # Access the fit before normalization
        gfit_before = rp.gaussian_fit
        gprof_before = rp.gaussian_profile
        fwhm_before = rp.gaussian_fwhm
        amp_before = gfit_before.amplitude.value

        rp.normalize()

        # The fit should be a new object (recomputed)
        gfit_after = rp.gaussian_fit
        assert gfit_after is not gfit_before

        # The amplitude should be close to 1 for the normalized profile
        assert gfit_after.amplitude.value == pytest.approx(1.0, abs=0.2)
        assert gfit_after.amplitude.value != pytest.approx(amp_before, rel=0.1)

        # The gaussian_profile should also be recomputed
        gprof_after = rp.gaussian_profile
        assert gprof_after is not gprof_before
        assert_allclose(gprof_after, gfit_after(rp.radius))

        # FWHM should be approximately the same (shape doesn't change)
        fwhm_after = rp.gaussian_fwhm
        assert_allclose(fwhm_after, fwhm_before, rtol=0.1)

    def test_moffat_fit_invalidated_on_normalize(self, profile_data):
        """
        Test that Moffat fit properties are invalidated when the profile
        is normalized, and the recomputed fit matches the normalized
        profile.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii)

        # Access the fit before normalization
        mfit_before = rp.moffat_fit
        mprof_before = rp.moffat_profile
        fwhm_before = rp.moffat_fwhm
        amp_before = mfit_before.amplitude.value

        rp.normalize()

        # The fit should be a new object (recomputed)
        mfit_after = rp.moffat_fit
        assert mfit_after is not mfit_before

        # The amplitude should differ from the unnormalized fit
        assert mfit_after.amplitude.value != pytest.approx(amp_before, rel=0.1)

        # The moffat_profile should also be recomputed
        mprof_after = rp.moffat_profile
        assert mprof_after is not mprof_before
        assert_allclose(mprof_after, mfit_after(rp.radius))

        # FWHM should be approximately the same (shape doesn't change)
        fwhm_after = rp.moffat_fwhm
        assert_allclose(fwhm_after, fwhm_before, rtol=0.1)

    def test_fit_invalidated_on_unnormalize(self, profile_data):
        """
        Test that Gaussian and Moffat fits are invalidated when
        unnormalize is called, and the recomputed fits match the
        original (unnormalized) profile.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii)

        # Get fits on the original profile
        gfit_orig = rp.gaussian_fit
        mfit_orig = rp.moffat_fit
        gfwhm_orig = rp.gaussian_fwhm
        mfwhm_orig = rp.moffat_fwhm

        # Normalize and access fits on the normalized profile
        rp.normalize()
        gfit_norm = rp.gaussian_fit
        mfit_norm = rp.moffat_fit
        assert gfit_norm is not gfit_orig
        assert mfit_norm is not mfit_orig

        # Unnormalize and verify fits are recomputed to match original
        rp.unnormalize()
        gfit_unnorm = rp.gaussian_fit
        mfit_unnorm = rp.moffat_fit

        # Should be new objects (not the cached normalized ones)
        assert gfit_unnorm is not gfit_norm
        assert mfit_unnorm is not mfit_norm

        # Amplitudes should match the original fits
        assert_allclose(gfit_unnorm.amplitude.value,
                        gfit_orig.amplitude.value, rtol=0.01)
        assert_allclose(mfit_unnorm.amplitude.value,
                        mfit_orig.amplitude.value, rtol=0.01)

        # FWHMs should match the originals
        assert_allclose(rp.gaussian_fwhm, gfwhm_orig, rtol=0.01)
        assert_allclose(rp.moffat_fwhm, mfwhm_orig, rtol=0.01)

    def test_fit_not_accessed_before_normalize(self, profile_data):
        """
        Test that fits computed after normalization (without prior
        access) correspond to the normalized profile.
        """
        xycen, data, _, _ = profile_data

        edge_radii = np.arange(36)
        rp = RadialProfile(data, xycen, edge_radii)

        # Normalize without ever accessing the fit first
        rp.normalize()

        # The fit should be on the normalized profile
        assert rp.gaussian_fit.amplitude.value == pytest.approx(1.0, abs=0.2)
        assert rp.moffat_fit.amplitude.value == pytest.approx(1.0, abs=0.2)
        assert rp.gaussian_profile is not None
        assert rp.moffat_profile is not None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for performing aperture photometry.
"""

import warnings

import astropy.units as u
import numpy as np
from astropy.nddata import NDData, StdDevUncertainty
from astropy.utils.exceptions import AstropyUserWarning

from photutils.aperture.converters import region_to_aperture
from photutils.aperture.core import Aperture, SkyAperture, _aperture_metadata
from photutils.utils._deprecation import (create_empty_deprecated_qtable,
                                          deprecated_positional_kwargs)
from photutils.utils._misc import _get_meta

__all__ = ['aperture_photometry']


# Remove in 4.0
_DEPRECATED_COLUMNS: dict = {
    'xcenter': 'x_center',
    'ycenter': 'y_center',
}


@deprecated_positional_kwargs(since='3.0', until='4.0')
def aperture_photometry(data, apertures, error=None, mask=None,
                        method='exact', subpixels=5, wcs=None):
    """
    Perform aperture photometry on the input data by summing the flux
    within the given aperture(s).

    Note that this function returns the sum of the (weighted) input
    ``data`` values within the aperture. It does not convert data
    in surface brightness units to flux or counts. Conversion from
    surface-brightness units should be performed before using this
    function.

    Parameters
    ----------
    data : array_like, `~astropy.units.Quantity`, `~astropy.nddata.NDData`
        The 2D array on which to perform photometry. ``data``
        should be background-subtracted. If ``data`` is a
        `~astropy.units.Quantity` array, then ``error`` (if input)
        must also be a `~astropy.units.Quantity` array with the same
        units. See the Notes section below for more information about
        `~astropy.nddata.NDData` input.

    apertures : `~photutils.aperture.Aperture`, supported `regions.Region`, \
        list of `~photutils.aperture.Aperture` or `regions.Region`
        The aperture(s) to use for the photometry. If ``apertures`` is
        a list of `~photutils.aperture.Aperture` or `regions.Region`,
        then they all must have the same position(s). If
        ``apertures`` contains a `~photutils.aperture.SkyAperture` or
        `~regions.SkyRegion` object, then a WCS must be input using
        the ``wcs`` keyword. Region objects are converted to aperture
        objects.

    error : array_like or `~astropy.units.Quantity`, optional
        The pixel-wise Gaussian 1-sigma errors of the input
        ``data``. ``error`` is assumed to include *all* sources
        of error, including the Poisson error of the sources (see
        `~photutils.utils.calc_total_error`). ``error`` must have the
        same shape as the input ``data``. If a `~astropy.units.Quantity`
        array, then ``data`` must also be a `~astropy.units.Quantity`
        array with the same units.

    mask : array_like (bool), optional
        A boolean mask with the same shape as ``data`` where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked data are excluded from all calculations.

    method : {'exact', 'center', 'subpixel'}, optional
        The method used to determine the overlap of the aperture on
        the pixel grid. Not all options are available for all aperture
        types. Note that the more precise methods are generally slower.
        The following methods are available:

        * ``'exact'`` (default):
            The exact fractional overlap of the aperture and each pixel
            is calculated. The aperture weights will contain values
            between 0 and 1.

        * ``'center'``:
            A pixel is considered to be entirely in or out of the
            aperture depending on whether its center is in or out of the
            aperture. The aperture weights will contain values only of 0
            (out) and 1 (in).

        * ``'subpixel'``:
            A pixel is divided into subpixels (see the ``subpixels``
            keyword), each of which are considered to be entirely in
            or out of the aperture depending on whether its center is
            in or out of the aperture. If ``subpixels=1``, this method
            is equivalent to ``'center'``. The aperture weights will
            contain values between 0 and 1.

    subpixels : int, optional
        For the ``'subpixel'`` method, resample pixels by this factor
        in each dimension. That is, each pixel is divided into
        ``subpixels**2`` subpixels. This keyword is ignored unless
        ``method='subpixel'``.

    wcs : WCS object, optional
        A world coordinate system (WCS) transformation that
        supports the `astropy shared interface for WCS
        <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_ (e.g.,
        `astropy.wcs.WCS`, `gwcs.wcs.WCS`). If provided, the output
        table will include a ``'sky_center'`` column with the sky
        coordinates of the input aperture center(s). This keyword is
        required if the input ``apertures`` contains a `SkyAperture` or
        `~regions.SkyRegion`.

    Returns
    -------
    table : `~astropy.table.QTable`
        A table of the photometry with the following columns:

        * ``'id'``:
          The source ID.

        * ``'x_center'``, ``'y_center'``:
          The ``x`` and ``y`` pixel coordinates of the input aperture
          center(s).

        * ``'sky_center'``:
          The sky coordinates of the input aperture center(s). Returned
          if a ``wcs`` is input.

        * ``'aperture_sum'``:
          The sum of the values within the aperture(s).

        * ``'aperture_sum_err'``:
          The corresponding uncertainty in the ``'aperture_sum'``
          values. Returned only if the input ``error`` is not `None`.

        The table metadata includes the Astropy and Photutils version
        numbers and the `aperture_photometry` calling arguments.

    Notes
    -----
    `~regions.Region` objects are converted to `Aperture` objects using
    the :func:`region_to_aperture` function.

    `RectangularAperture` and `RectangularAnnulus` photometry with the
    "exact" method uses a subpixel approximation by subdividing each
    data pixel by a factor of 1024 (``subpixels = 32``). For rectangular
    aperture widths and heights in the range from 2 to 100 pixels, this
    subpixel approximation gives results typically within 0.001 percent
    or better of the exact value. The differences can be larger for
    smaller apertures (e.g., aperture sizes of one pixel or smaller).
    For such small sizes, it is recommended to set ``method='subpixel'``
    with a larger ``subpixels`` size.

    If the input ``data`` is a `~astropy.nddata.NDData` instance,
    then the ``error``, ``mask``, and ``wcs`` keyword inputs are
    ignored. Instead, these values should be defined as attributes in
    the `~astropy.nddata.NDData` object. In the case of ``error``,
    it must be defined in the ``uncertainty`` attribute with a
    `~astropy.nddata.StdDevUncertainty` instance.
    """
    if isinstance(data, NDData):
        nddata_attr = {'error': error, 'mask': mask, 'wcs': wcs}
        for key, value in nddata_attr.items():
            if value is not None:
                msg = (f'The {key!r} keyword is ignored. Its value '
                       'is obtained from the input NDData object.')
                warnings.warn(msg, AstropyUserWarning)

        mask = data.mask
        wcs = data.wcs

        if isinstance(data.uncertainty, StdDevUncertainty):
            if data.uncertainty.unit is None:
                error = data.uncertainty.array
            else:
                error = data.uncertainty.array * data.uncertainty.unit

        if data.unit is not None:
            data = u.Quantity(data.data, unit=data.unit)
        else:
            data = data.data

        return aperture_photometry(data, apertures, error=error, mask=mask,
                                   method=method, subpixels=subpixels,
                                   wcs=wcs)

    single_aperture = False
    if not isinstance(apertures, (list, tuple, np.ndarray)):
        single_aperture = True
        apertures = (apertures,)

    # Create table metadata using the input apertures, not the converted
    # ones
    aper_meta = {}
    for i, aperture in enumerate(apertures):
        i = '' if single_aperture else i
        aper_meta.update(_aperture_metadata(aperture, index=i))

    # Convert regions to apertures if necessary
    apertures = [region_to_aperture(aper)
                 if not isinstance(aper, Aperture) else aper
                 for aper in apertures]

    # Convert sky to pixel apertures
    skyaper = False
    if isinstance(apertures[0], SkyAperture):
        if wcs is None:
            msg = ('A WCS transform must be defined by the input data or '
                   'the wcs keyword when using a SkyAperture object.')
            raise ValueError(msg)

        # Include SkyCoord position in the output table
        skyaper = True
        skycoord_pos = apertures[0].positions

        apertures = [aper.to_pixel(wcs) for aper in apertures]

    # Compare positions in pixels to avoid comparing SkyCoord objects
    positions = apertures[0].positions
    for aper in apertures[1:]:
        if not np.array_equal(aper.positions, positions):
            msg = 'Input apertures must all have identical positions'
            raise ValueError(msg)

    # Define output table meta data
    meta = _get_meta()
    calling_args = f"method='{method}', subpixels={subpixels}"
    meta['aperture_photometry_args'] = calling_args
    meta.update(aper_meta)

    # Replace with QTable in 4.0
    tbl = create_empty_deprecated_qtable(
        _DEPRECATED_COLUMNS, since='3.0', until='4.0')

    tbl.meta.update(meta)  # keep tbl.meta type

    positions = np.atleast_2d(apertures[0].positions)
    tbl['id'] = np.arange(positions.shape[0], dtype=int) + 1

    xypos_pixel = np.transpose(positions)
    tbl['x_center'] = xypos_pixel[0]
    tbl['y_center'] = xypos_pixel[1]

    if skyaper:
        if skycoord_pos.isscalar:
            # Create length-1 SkyCoord array
            tbl['sky_center'] = skycoord_pos.reshape((-1,))
        else:
            tbl['sky_center'] = skycoord_pos

    if wcs is not None and not skyaper:
        tbl['sky_center'] = wcs.pixel_to_world(*np.transpose(positions))

    sum_key_main = 'aperture_sum'
    sum_err_key_main = 'aperture_sum_err'
    for i, aper in enumerate(apertures):
        aper_sum, aper_sum_err = aper.do_photometry(data, error=error,
                                                    mask=mask, method=method,
                                                    subpixels=subpixels)

        sum_key = sum_key_main
        sum_err_key = sum_err_key_main
        if not single_aperture:
            sum_key += f'_{i}'
            sum_err_key += f'_{i}'

        tbl[sum_key] = aper_sum
        if error is not None:
            tbl[sum_err_key] = aper_sum_err

    return tbl



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for reading and writing PSF models.
"""

import io
import itertools
import os
import warnings

import numpy as np
from astropy.io import fits, registry
from astropy.io.fits.verify import VerifyWarning
from astropy.nddata import NDData, reshape_as_blocks

from photutils.utils._deprecation import deprecated_positional_kwargs

__all__ = ['GriddedPSFModelRead', 'stdpsf_reader', 'webbpsf_reader']
__doctest_skip__ = ['GriddedPSFModelRead']


class GriddedPSFModelRead(registry.UnifiedReadWrite):
    """
    Read and parse a FITS file into a `GriddedPSFModel` instance.

    This class enables the astropy unified I/O layer for
    `~photutils.psf.GriddedPSFModel`. This allows easily reading a file
    in different supported data formats using syntax such as::

      >>> from photutils.psf import GriddedPSFModel
      >>> psf_model = GriddedPSFModel.read('filename.fits', format=format)

    Get help on the available readers for
    `~photutils.psf.GriddedPSFModel` using the ``help()`` method::

      >>> # Get help reading Table and list supported formats
      >>> GriddedPSFModel.read.help()

      >>> # Get detailed help on the STSPSF FITS reader
      >>> GriddedPSFModel.read.help('stdpsf')

      >>> # Get detailed help on the WebbPSF FITS reader
      >>> GriddedPSFModel.read.help('webbpsf')

      >>> # Print list of available formats
      >>> GriddedPSFModel.read.list_formats()

    Parameters
    ----------
    instance : object
        Descriptor calling instance or `None` if no instance.

    cls : type
        Descriptor calling class (either owner class or instance class).
    """

    def __init__(self, instance, cls):
        # uses default global registry
        super().__init__(instance, cls, 'read', registry=None)

    def __call__(self, *args, **kwargs):
        """
        Read and parse a FITS file into a `GriddedPSFModel` instance
        using the registered "read" function.

        Parameters
        ----------
        *args : tuple
            Positional arguments passed through to data reader. The
            first argument is typically the input filename.

        **kwargs : dict, optional
            Keyword arguments passed through to data reader. This
            includes the ``format`` keyword argument.

        Returns
        -------
        out : `~photutils.psf.GriddedPSFModel`
            A gridded ePSF model corresponding to FITS file contents.
        """
        return self.registry.read(self._cls, *args, **kwargs)


def _read_stdpsf(filename):
    """
    Read a STScI standard-format ePSF (STDPSF) FITS file.

    Parameters
    ----------
    filename : str
        The name of the STDPDF FITS file.

    Returns
    -------
    data : dict
        A dictionary containing the ePSF data and metadata.
    """
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', VerifyWarning)
        with fits.open(filename, ignore_missing_end=True) as hdulist:
            header = hdulist[0].header
            data = hdulist[0].data

    try:
        npsfs = header['NAXIS3']
        nxpsfs = header['NXPSFS']
        nypsfs = header['NYPSFS']
    except KeyError as exc:
        msg = 'Invalid STDPDF FITS file'
        raise ValueError(msg) from exc

    if 'IPSFX01' in header:
        xgrid = [header[f'IPSFX{i:02d}'] for i in range(1, nxpsfs + 1)]
        ygrid = [header[f'JPSFY{i:02d}'] for i in range(1, nypsfs + 1)]
    elif 'IPSFXA5' in header:
        xgrid = []
        ygrid = []
        xkeys = ('IPSFXA5', 'IPSFXB5', 'IPSFXC5', 'IPSFXD5')
        for xkey in xkeys:
            xgrid.extend([int(n) for n in header[xkey].split()])
        ykeys = ('JPSFYA5', 'JPSFYB5')
        for ykey in ykeys:
            ygrid.extend([int(n) for n in header[ykey].split()])
    else:
        msg = 'Unknown STDPSF FITS file'
        raise ValueError(msg)

    # STDPDF FITS positions are 1-indexed
    xgrid = np.array(xgrid) - 1
    ygrid = np.array(ygrid) - 1

    # nypsfs, nxpsfs, detector
    # 6, 6     WFPC2, 4 det
    # 1, 1     ACS/HRC
    # 10, 9    ACS/WFC, 2 det
    # 3, 3     WFC3/IR
    # 8, 7     WFC3/UVIS, 2 det
    # 5, 5     NIRISS
    # 5, 5     NIRCam SW
    # 10, 20   NIRCam SW (NRCSW), 8 det
    # 5, 5     NIRCam LW
    # 3, 3     MIRI

    return {'data': data,
            'npsfs': npsfs,
            'nxpsfs': nxpsfs,
            'nypsfs': nypsfs,
            'xgrid': xgrid,
            'ygrid': ygrid}


def _split_detectors(grid_data, detector_data, detector_id):
    """
    Split an ePSF array into individual detectors.

    Parameters
    ----------
    grid_data : dict
        A dictionary containing the ePSF data and metadata.

    detector_data : dict
        A dictionary containing the detector data.

    detector_id : int
        The detector ID.

    Returns
    -------
    data : `~numpy.ndarray`
        The ePSF data for the specified detector.

    xgrid : `~numpy.ndarray`
        The x-grid for the specified detector.

    ygrid : `~numpy.ndarray`
        The y-grid for the specified detector.

    Notes
    -----
    In particular::

    * HST WFPC2 STDPSF file contains 4 detectors
    * HST ACS/WFC STDPSF file contains 2 detectors
    * HST WFC3/UVIS STDPSF file contains 2 detectors
    * JWST NIRCam "NRCSW" STDPSF file contains 8 detectors
    """
    data = grid_data['data']
    npsfs = grid_data['npsfs']
    nxpsfs = grid_data['nxpsfs']
    nypsfs = grid_data['nypsfs']
    xgrid = grid_data['xgrid']
    ygrid = grid_data['ygrid']
    nxdet = detector_data['nxdet']
    nydet = detector_data['nydet']
    det_map = detector_data['det_map']
    det_size = detector_data['det_size']

    ii = np.arange(npsfs).reshape((nypsfs, nxpsfs))
    nxpsfs //= nxdet
    nypsfs //= nydet
    ndet = nxdet * nydet
    ii = reshape_as_blocks(ii, (nypsfs, nxpsfs))
    ii = ii.reshape(ndet, npsfs // ndet)

    # detector_id -> index
    det_idx = det_map[detector_id]
    idx = ii[det_idx]
    data = data[idx]

    xp = det_idx % nxdet
    i0 = xp * nxpsfs
    i1 = i0 + nxpsfs
    xgrid = xgrid[i0:i1] - xp * det_size

    ygrid = ygrid[:nypsfs] if det_idx < nxdet else ygrid[nypsfs:] - det_size

    return data, xgrid, ygrid


def _split_wfc_uvis(grid_data, detector_id):
    """
    Split an ePSF array into individual WFC/UVIS detectors.

    Parameters
    ----------
    grid_data : dict
        A dictionary containing the ePSF data and metadata.

    detector_id : int
        The detector ID.

    Returns
    -------
    data : `~numpy.ndarray`
        The ePSF data for the specified detector.

    xgrid : `~numpy.ndarray`
        The x-grid for the specified detector.

    ygrid : `~numpy.ndarray`
        The y-grid for the specified detector.
    """
    if detector_id is None:
        msg = 'detector_id must be specified for ACS/WFC and WFC3/UVIS ePSFs'
        raise ValueError(msg)
    if detector_id not in (1, 2):
        msg = 'detector_id must be 1 or 2'
        raise ValueError(msg)

    # ACS/WFC1 and WFC3/UVIS1 chip1 (sci, 2) are above chip2 (sci, 1)
    # in y-pixel coordinates
    xgrid = grid_data['xgrid']
    ygrid = grid_data['ygrid']
    ygrid = ygrid.reshape((2, ygrid.shape[0] // 2))[detector_id - 1]
    if detector_id == 2:
        ygrid -= 2048

    npsfs = grid_data['npsfs']
    data = grid_data['data']
    data_ny, data_nx = data.shape[1:]
    data = data.reshape((2, npsfs // 2, data_ny, data_nx))[detector_id - 1]

    return data, xgrid, ygrid


def _split_wfpc2(grid_data, detector_id):
    """
    Split an ePSF array into individual WFPC2 detectors.

    Parameters
    ----------
    grid_data : dict
        A dictionary containing the ePSF data and metadata.

    detector_id : int
        The detector ID.

    Returns
    -------
    data : `~numpy.ndarray`
        The ePSF data for the specified detector.

    xgrid : `~numpy.ndarray`
        The x-grid for the specified detector.

    ygrid : `~numpy.ndarray`
        The y-grid for the specified detector.
    """
    if detector_id is None:
        msg = 'detector_id must be specified for WFPC2 ePSFs'
        raise ValueError(msg)
    if detector_id not in range(1, 5):
        msg = 'detector_id must be between 1 and 4, inclusive'
        raise ValueError(msg)

    nxdet = 2
    nydet = 2
    det_size = 800

    # det (exten:idx)
    # WF2 (2:2)  PC (1:3)
    # WF3 (3:0)  WF4 (4:1)
    det_map = {1: 3, 2: 2, 3: 0, 4: 1}

    detector_data = {'nxdet': nxdet,
                     'nydet': nydet,
                     'det_size': det_size,
                     'det_map': det_map}

    return _split_detectors(grid_data, detector_data, detector_id)


def _split_nrcsw(grid_data, detector_id):
    """
    Split an ePSF array into individual NIRCam SW detectors.

    Parameters
    ----------
    grid_data : dict
        A dictionary containing the ePSF data and metadata.

    detector_id : int
        The detector ID.

    Returns
    -------
    data : `~numpy.ndarray`
        The ePSF data for the specified detector.

    xgrid : `~numpy.ndarray`
        The x-grid for the specified detector.

    ygrid : `~numpy.ndarray`
        The y-grid for the specified detector.
    """
    if detector_id is None:
        msg = 'detector_id must be specified for NRCSW ePSFs'
        raise ValueError(msg)
    if detector_id not in range(1, 9):
        msg = 'detector_id must be between 1 and 8, inclusive'
        raise ValueError(msg)

    nxdet = 4
    nydet = 2
    det_size = 2048

    # det (ext:idx)
    # A2 (2:4)  A4 (4:5)  B3 (7:6)  B1 (5:7)
    # A1 (1:0)  A3 (3:1)  B4 (8:2)  B2 (6:3)
    det_map = {1: 0, 3: 1, 8: 2, 6: 3, 2: 4, 4: 5, 7: 6, 5: 7}

    detector_data = {'nxdet': nxdet,
                     'nydet': nydet,
                     'det_size': det_size,
                     'det_map': det_map}

    return _split_detectors(grid_data, detector_data, detector_id)


def _get_metadata(filename, detector_id):
    """
    Get metadata from the filename and ``detector_id``.

    Parameters
    ----------
    filename : str
        The name of the STDPDF FITS file.

    detector_id : int
        The detector ID.

    Returns
    -------
    meta : dict or `None`
        A dictionary containing the metadata.
    """
    if isinstance(filename, io.FileIO):
        filename = filename.name

    parts = os.path.basename(filename).strip('.fits').split('_')
    if len(parts) not in (3, 4):
        return None  # filename from astropy download_file

    detector, filter_name = parts[1:3]
    meta = {'STDPSF': filename,
            'detector': detector,
            'filter': filter_name}

    if detector_id is not None:
        detector_map = {'WFPC2': ['HST/WFPC2', 'WFPC2'],
                        'ACSHRC': ['HST/ACS', 'HRC'],
                        'ACSWFC': ['HST/ACS', 'WFC'],
                        'WFC3UV': ['HST/WFC3', 'UVIS'],
                        'WFC3IR': ['HST/WFC3', 'IR'],
                        'NRCSW': ['JWST/NIRCam', 'NRCSW'],
                        'NRCA1': ['JWST/NIRCam', 'A1'],
                        'NRCA2': ['JWST/NIRCam', 'A2'],
                        'NRCA3': ['JWST/NIRCam', 'A3'],
                        'NRCA4': ['JWST/NIRCam', 'A4'],
                        'NRCB1': ['JWST/NIRCam', 'B1'],
                        'NRCB2': ['JWST/NIRCam', 'B2'],
                        'NRCB3': ['JWST/NIRCam', 'B3'],
                        'NRCB4': ['JWST/NIRCam', 'B4'],
                        'NRCAL': ['JWST/NIRCam', 'A5'],
                        'NRCBL': ['JWST/NIRCam', 'B5'],
                        'NIRISS': ['JWST/NIRISS', 'NIRISS'],
                        'MIRI': ['JWST/MIRI', 'MIRIM']}

        try:
            inst_det = detector_map[detector]
        except KeyError as exc:
            msg = f'Unknown detector {detector}'
            raise ValueError(msg) from exc

        if inst_det[1] == 'WFPC2':
            wfpc2_map = {1: 'PC', 2: 'WF2', 3: 'WF3', 4: 'WF4'}
            inst_det[1] = wfpc2_map[detector_id]

        if inst_det[1] in ('WFC', 'UVIS'):
            chip = 2 if detector_id == 1 else 1
            inst_det[1] = f'{inst_det[1]}{chip}'

        if inst_det[1] == 'NRCSW':
            sw_map = {1: 'A1', 2: 'A2', 3: 'A3', 4: 'A4',
                      5: 'B1', 6: 'B2', 7: 'B3', 8: 'B4'}
            inst_det[1] = sw_map[detector_id]

        meta['instrument'] = inst_det[0]
        meta['detector'] = inst_det[1]

    return meta


@deprecated_positional_kwargs(since='3.0', until='4.0')
def stdpsf_reader(filename, detector_id=None):
    """
    Generate a `~photutils.psf.GriddedPSFModel` from a STScI standard-
    format ePSF (STDPSF) FITS file.

    .. note::
        Instead of being used directly, this function is intended to
        be used via the `~photutils.psf.GriddedPSFModel` ``read``
        method, e.g., ``model = GriddedPSFModel.read(filename,
        format='stdpsf')``.

    STDPSF files are FITS files that contain a 3D array of ePSFs with
    the header detailing where the fiducial ePSFs are located in the
    detector coordinate frame.

    The oversampling factor for STDPSF FITS files is assumed to be 4.

    Parameters
    ----------
    filename : str
        The name of the STDPDF FITS file. A URL can also be used.

    detector_id : `None` or int, optional
        For STDPSF files that contain ePSF grids for multiple detectors,
        one will need to identify the detector for which to extract the
        ePSF grid. This keyword is ignored for STDPSF files that do not
        contain ePSF grids for multiple detectors.

        For WFPC2, the detector value (int) should be:

        * 1: PC, 2: WF2, 3: WF3, 4: WF4

        For ACS/WFC and WFC3/UVIS, the detector value should be:

        * 1: WFC2, UVIS2 (sci, 1)
        * 2: WFC1, UVIS1 (sci, 2)

        Note that for these two instruments, detector 1 is above
        detector 2 in the y direction. However, in the FLT FITS files,
        the (sci, 1) extension corresponds to detector 2 (WFC2, UVIS2)
        and the (sci, 2) extension corresponds to detector 1 (WFC1,
        UVIS1).

        For NIRCam NRCSW files that contain ePSF grids for all 8 SW
        detectors, the detector value should be:

        * 1: A1, 2: A2, 3: A3, 4: A4
        * 5: B1, 6: B2, 7: B3, 8: B4

    Returns
    -------
    model : `~photutils.psf.GriddedPSFModel`
        The gridded ePSF model.
    """
    from photutils.psf import GriddedPSFModel  # prevent circular import

    grid_data = _read_stdpsf(filename)

    npsfs = grid_data['npsfs']
    if npsfs in (90, 56, 36, 200):
        if npsfs in (90, 56):  # ACS/WFC or WFC3/UVIS data (2 chips)
            data, xgrid, ygrid = _split_wfc_uvis(grid_data, detector_id)
        elif npsfs == 36:  # WFPC2 data (4 chips)
            data, xgrid, ygrid = _split_wfpc2(grid_data, detector_id)
        elif npsfs == 200:  # NIRCam SW data (8 chips)
            data, xgrid, ygrid = _split_nrcsw(grid_data, detector_id)
        else:
            msg = 'Unknown detector or STDPSF format'
            raise ValueError(msg)
    else:
        data = grid_data['data']
        xgrid = grid_data['xgrid']
        ygrid = grid_data['ygrid']

    # itertools.product iterates over the last input first
    xy_grid = [yx[::-1] for yx in itertools.product(ygrid, xgrid)]

    oversampling = 4  # assumption for STDPSF files
    nxpsfs = xgrid.shape[0]
    nypsfs = ygrid.shape[0]
    meta = {'grid_xypos': xy_grid,
            'oversampling': oversampling,
            'nxpsfs': nxpsfs,
            'nypsfs': nypsfs}

    # try to get additional metadata from the filename because this
    # information is not currently available in the FITS headers
    file_meta = _get_metadata(filename, detector_id)
    if file_meta is not None:
        meta.update(file_meta)

    return GriddedPSFModel(NDData(data, meta=meta))


def webbpsf_reader(filename):
    """
    Generate a `~photutils.psf.GriddedPSFModel` from a WebbPSF FITS file
    containing a PSF grid.

    .. note::
        Instead of being used directly, this function is intended to
        be used via the `~photutils.psf.GriddedPSFModel` ``read``
        method, e.g., ``model = GriddedPSFModel.read(filename,
        format='webbpsf')``.

    The WebbPSF FITS file contain a 3D array of ePSFs with the header
    detailing where the fiducial ePSFs are located in the detector
    coordinate frame.

    Parameters
    ----------
    filename : str
        The name of the WebbPSF FITS file. A URL can also be used.

    Returns
    -------
    model : `~photutils.psf.GriddedPSFModel`
        The gridded ePSF model.
    """
    from photutils.psf import GriddedPSFModel  # prevent circular import

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', VerifyWarning)
        with fits.open(filename, ignore_missing_end=True) as hdulist:
            header = hdulist[0].header
            data = hdulist[0].data

    # handle the case of only one 2D PSF
    data = np.atleast_3d(data)

    if not any('DET_YX' in key for key in header):
        msg = 'Invalid WebbPSF FITS file; missing "DET_YX{}" header keys'
        raise ValueError(msg)
    if 'OVERSAMP' not in header:
        msg = 'Invalid WebbPSF FITS file; missing "OVERSAMP" header key'
        raise ValueError(msg)

    # convert header to meta dict
    header = header.copy(strip=True)
    header.pop('HISTORY', None)
    header.pop('COMMENT', None)
    header.pop('', None)
    meta = dict(header)
    meta = {key.lower(): meta[key] for key in meta}  # user lower-case keys

    # define grid_xypos from DET_YX{} FITS header keywords
    xypos = []
    for key in meta:
        if 'det_yx' in key:
            vals = header[key].lstrip('(').rstrip(')').split(',')
            xypos.append((float(vals[0]), float(vals[1])))
    meta['grid_xypos'] = xypos

    if 'oversampling' not in meta:
        meta['oversampling'] = meta['oversamp']

    ndd = NDData(data, meta=meta)

    return GriddedPSFModel(ndd)


def is_stdpsf(origin, filepath, fileobj, *args, **kwargs):
    """
    Determine whether a file is a STDPSF FITS file.

    Parameters
    ----------
    origin : {'read', 'write'}
        A string indicating whether the file is to be opened for reading
        or writing.

    filepath : str
        The file path of the FITS file.

    fileobj : file-like object
        An open file object to read the file's contents, or `None` if
        the file could not be opened.

    *args, **kwargs
        Any additional positional or keyword arguments for the read or
        write function.

    Returns
    -------
    result : bool
        Returns `True` if the given file is a STDPSF FITS file.
    """
    if filepath is not None:
        extens = ('.fits', '.fits.gz', '.fit', '.fit.gz', '.fts', '.fts.gz')
        isfits = filepath.lower().endswith(extens)
        if isfits:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', VerifyWarning)
                header = fits.getheader(filepath)
            keys = ('NAXIS3', 'NXPSFS', 'NYPSFS')
            return all(key in header for key in keys)

    return False


def is_webbpsf(origin, filepath, fileobj, *args, **kwargs):
    """
    Determine whether a file is a WebbPSF FITS file.

    Parameters
    ----------
    origin : {'read', 'write'}
        A string indicating whether the file is to be opened for reading
        or writing.

    filepath : str
        The file path of the FITS file.

    fileobj : file-like object
        An open file object to read the file's contents, or `None` if
        the file could not be opened.

    *args, **kwargs
        Any additional positional or keyword arguments for the read or
        write function.

    Returns
    -------
    result : bool
        Returns `True` if the given file is a WebbPSF FITS file.
    """
    if filepath is not None:
        extens = ('.fits', '.fits.gz', '.fit', '.fit.gz', '.fts', '.fts.gz')
        isfits = filepath.lower().endswith(extens)
        if isfits:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', VerifyWarning)
                header = fits.getheader(filepath)
            keys = ('NAXIS3', 'OVERSAMP', 'DET_YX0')
            return all(key in header for key in keys)

    return False



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for generating matplotlib colormaps.

This module requires matplotlib to be installed.
"""

import numpy as np

from photutils.utils._deprecation import (deprecated_positional_kwargs,
                                          deprecated_renamed_argument)

__all__ = ['make_random_cmap']


@deprecated_renamed_argument('ncolors', 'n_colors', '3.0', until='4.0')
@deprecated_positional_kwargs(since='3.0', until='4.0')
def make_random_cmap(n_colors=256, seed=None):
    """
    Make a matplotlib colormap consisting of (random) muted colors.

    A random colormap is very useful for plotting segmentation images.

    Parameters
    ----------
    n_colors : int, optional
        The number of colors in the colormap. The default is 256. Must
        be at least 1.

    seed : int, optional
        A seed to initialize the `numpy.random.BitGenerator`. If `None`,
        then fresh, unpredictable entropy will be pulled from the OS.
        Separate function calls with the same ``seed`` will generate the
        same colormap.

    Returns
    -------
    cmap : `matplotlib.colors.ListedColormap`
        The matplotlib colormap with random colors in RGBA format.
    """
    if n_colors < 1:
        msg = 'n_colors must be at least 1'
        raise ValueError(msg)

    from matplotlib import colors

    rng = np.random.default_rng(seed)
    hue = rng.uniform(low=0.0, high=1.0, size=n_colors)
    sat = rng.uniform(low=0.2, high=0.7, size=n_colors)
    val = rng.uniform(low=0.5, high=1.0, size=n_colors)
    hsv = np.dstack((hue, sat, val))
    rgb = np.squeeze(colors.hsv_to_rgb(hsv))

    return colors.ListedColormap(colors.to_rgba_array(rgb))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the attributes module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord

from photutils.aperture.attributes import (ApertureAttribute, PixelPositions,
                                           PositiveScalar, PositiveScalarAngle,
                                           ScalarAngle, ScalarAngleOrValue,
                                           SkyCoordPositions)


class MockBase:
    """
    Minimal class using ApertureAttribute as a descriptor.
    """

    attr = ApertureAttribute(doc='a test attribute')


class MockWithLazy:
    """
    Minimal class with _lazyproperties for reset testing.
    """

    attr = ApertureAttribute()
    _lazyproperties = ['cached']  # noqa: RUF012


class MockPixelPos:
    """
    Minimal class using PixelPositions as a descriptor.
    """

    positions = PixelPositions()


class MockSkyCoord:
    """
    Minimal class using SkyCoordPositions as a descriptor.
    """

    positions = SkyCoordPositions()


class MockPositiveScalar:
    """
    Minimal class using PositiveScalar as a descriptor.
    """

    r = PositiveScalar()


class MockScalarAngle:
    """
    Minimal class using ScalarAngle as a descriptor.
    """

    angle = ScalarAngle()


class MockPositiveScalarAngle:
    """
    Minimal class using PositiveScalarAngle as a descriptor.
    """

    angle = PositiveScalarAngle()


class MockScalarAngleOrValue:
    """
    Minimal class using ScalarAngleOrValue as a descriptor.
    """

    theta = ScalarAngleOrValue()


class TestApertureAttribute:
    """
    Tests for the ApertureAttribute base descriptor class.
    """

    def test_class_access(self):
        """
        Test that __get__ with instance=None returns the descriptor.
        """
        desc = MockBase.attr
        assert isinstance(desc, ApertureAttribute)

    def test_doc(self):
        """
        Test that the descriptor docstring is set correctly.
        """
        desc = MockBase.__dict__['attr']
        assert desc.__doc__ == 'a test attribute'

    def test_set_name(self):
        """
        Test that __set_name__ assigns the attribute name correctly.
        """
        desc = MockBase.__dict__['attr']
        assert desc.name == 'attr'

    def test_set_and_get(self):
        """
        Test that setting and getting a value round-trips correctly.
        """
        obj = MockBase()
        obj.attr = 5.0
        assert obj.attr == 5.0

    def test_converts_to_float(self):
        """
        Test that non-Quantity scalars are converted to float.
        """
        obj = MockBase()
        obj.attr = 3
        assert isinstance(obj.attr, float)
        assert obj.attr == 3.0

    def test_quantity_not_converted(self):
        """
        Test that Quantity values are stored without conversion.
        """
        obj = MockBase()
        val = 5.0 * u.m
        obj.attr = val
        assert isinstance(obj.attr, u.Quantity)

    def test_skycoord_not_converted(self):
        """
        Test that SkyCoord values are stored without conversion.
        """
        obj = MockBase()
        val = SkyCoord(ra=10.0, dec=20.0, unit='deg')
        obj.attr = val
        assert isinstance(obj.attr, SkyCoord)

    def test_reset_lazyproperties(self):
        """
        Test that lazyproperties are cleared when the attribute is
        updated.
        """
        obj = MockWithLazy()
        obj.attr = 1.0
        obj.cached = 42
        assert 'cached' in obj.__dict__
        obj.attr = 2.0  # second assignment triggers _reset_lazyproperties
        assert 'cached' not in obj.__dict__

    def test_reset_no_lazyproperties(self):
        """
        Test that AttributeError is silently caught when _lazyproperties
        is absent.
        """
        obj = MockBase()
        obj.attr = 1.0
        # Triggers _reset_lazyproperties; obj has no _lazyproperties
        obj.attr = 2.0
        assert obj.attr == 2.0

    def test_delete(self):
        """
        Test that deleting the attribute removes it from the instance.
        """
        obj = MockBase()
        obj.attr = 5.0
        del obj.attr
        assert 'attr' not in obj.__dict__

    def test_validate_noop(self):
        """
        Test that the base _validate method is a no-op.
        """
        obj = MockBase()
        obj.attr = 99.0  # calls _validate internally without error
        assert obj.attr == 99.0


class TestPixelPositions:
    """
    Tests for the PixelPositions descriptor class.
    """

    def test_single_tuple(self):
        """
        Test that a single (x, y) tuple is accepted.
        """
        obj = MockPixelPos()
        obj.positions = (10, 20)
        # _validate returns the original (non-atleast_2d) array
        assert obj.positions.shape == (2,)
        assert obj.positions[0] == 10.0
        assert obj.positions[1] == 20.0

    def test_list_of_tuples(self):
        """
        Test that a list of (x, y) tuples is accepted.
        """
        obj = MockPixelPos()
        obj.positions = [(1, 2), (3, 4)]
        assert obj.positions.shape == (2, 2)

    def test_ndarray(self):
        """
        Test that a 2D numpy array of positions is accepted.
        """
        obj = MockPixelPos()
        pos = np.array([(5.0, 6.0), (7.0, 8.0)])
        obj.positions = pos
        assert obj.positions.shape == (2, 2)

    def test_zip_input(self):
        """
        Test that a zip of x and y arrays is accepted.
        """
        obj = MockPixelPos()
        obj.positions = zip([1.0, 2.0], [3.0, 4.0], strict=False)
        assert obj.positions.shape == (2, 2)

    def test_reset_lazyproperties(self):
        """
        Test that setting positions twice clears cached lazyproperties.
        """
        obj = MockPixelPos()
        obj.positions = [(1, 2)]
        obj._lazyproperties = ['cached']
        obj.cached = 99
        obj.positions = [(3, 4)]  # triggers reset
        assert 'cached' not in obj.__dict__

    def test_quantity_error(self):
        """
        Test that a Quantity array raises TypeError.
        """
        obj = MockPixelPos()
        match = "'positions' must not be a Quantity"
        with pytest.raises(TypeError, match=match):
            obj.positions = np.array([1.0, 2.0]) * u.m

    def test_bad_type_error(self):
        """
        Test the TypeError path via np.asanyarray(...).astype(float).
        """
        desc = MockPixelPos.__dict__['positions']
        match = "'positions' must not be a Quantity"
        with pytest.raises(TypeError, match=match):
            # Sets cannot be converted to float, triggering except TypeError
            desc._validate([({1, 2}, {3, 4})])

    def test_zip_quantity_error(self):
        """
        Test that TypeError is raised when zip contains Quantity objects.
        """
        obj = MockPixelPos()
        match = "'positions' must not be a Quantity"
        with pytest.raises(TypeError, match=match):
            obj.positions = zip(
                [1.0 * u.m, 2.0 * u.m],
                [3.0 * u.m, 4.0 * u.m],
                strict=False,
            )

    def test_nonfinite_nan_error(self):
        """
        Test that NaN positions raise ValueError.
        """
        obj = MockPixelPos()
        match = "'positions' must not contain any non-finite"
        with pytest.raises(ValueError, match=match):
            obj.positions = [(np.nan, 2.0)]

    def test_nonfinite_inf_error(self):
        """
        Test that infinite positions raise ValueError.
        """
        obj = MockPixelPos()
        match = "'positions' must not contain any non-finite"
        with pytest.raises(ValueError, match=match):
            obj.positions = [(1.0, np.inf)]

    def test_3d_error(self):
        """
        Test that a 3D array raises ValueError.
        """
        obj = MockPixelPos()
        match = "'positions' must be a"
        with pytest.raises(ValueError, match=match):
            obj.positions = np.ones((2, 2, 2))

    def test_wrong_ncols_error(self):
        """
        Test that positions with wrong column count raise ValueError.
        """
        obj = MockPixelPos()
        match = "'positions' must be a"
        with pytest.raises(ValueError, match=match):
            obj.positions = [(1.0, 2.0, 3.0), (4.0, 5.0, 6.0)]


class TestSkyCoordPositions:
    """
    Tests for the SkyCoordPositions descriptor class.
    """

    def test_valid(self):
        """
        Test that a SkyCoord value is accepted.
        """
        obj = MockSkyCoord()
        sky = SkyCoord(ra=10.0, dec=20.0, unit='deg')
        obj.positions = sky
        assert isinstance(obj.positions, SkyCoord)

    def test_invalid_type_error(self):
        """
        Test that a non-SkyCoord value raises TypeError.
        """
        obj = MockSkyCoord()
        match = "'positions' must be a SkyCoord instance"
        with pytest.raises(TypeError, match=match):
            obj.positions = (10.0, 20.0)


class TestPositiveScalar:
    """
    Tests for the PositiveScalar descriptor class.
    """

    def test_valid(self):
        """
        Test that a positive scalar value is accepted.
        """
        obj = MockPositiveScalar()
        obj.r = 3.5
        assert obj.r == 3.5

    def test_zero_error(self):
        """
        Test that zero raises ValueError.
        """
        obj = MockPositiveScalar()
        match = "'r' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            obj.r = 0.0

    def test_negative_error(self):
        """
        Test that a negative value raises ValueError.
        """
        obj = MockPositiveScalar()
        match = "'r' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            obj.r = -1.0

    def test_array_error(self):
        """
        Test that a non-scalar array raises ValueError.
        """
        obj = MockPositiveScalar()
        match = "'r' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            obj.r = np.array([1.0, 2.0])


class TestScalarAngle:
    """
    Tests for the ScalarAngle descriptor class.
    """

    def test_valid_rad(self):
        """
        Test that a scalar radian Quantity is accepted.
        """
        obj = MockScalarAngle()
        obj.angle = 0.5 * u.rad
        assert obj.angle == 0.5 * u.rad

    def test_valid_arcsec(self):
        """
        Test that a scalar arcsec Quantity is accepted.
        """
        obj = MockScalarAngle()
        obj.angle = 30.0 * u.arcsec
        assert obj.angle.unit == u.arcsec

    def test_nonscalar_error(self):
        """
        Test that a non-scalar Quantity raises ValueError.
        """
        obj = MockScalarAngle()
        match = "'angle' must be a scalar"
        with pytest.raises(ValueError, match=match):
            obj.angle = np.array([1.0, 2.0]) * u.arcsec

    def test_non_angle_units_error(self):
        """
        Test that a Quantity with non-angular units raises ValueError.
        """
        obj = MockScalarAngle()
        match = "'angle' must have angular units"
        with pytest.raises(ValueError, match=match):
            obj.angle = 5.0 * u.m

    def test_non_quantity_error(self):
        """
        Test that a plain float raises TypeError.
        """
        obj = MockScalarAngle()
        match = "'angle' must be a scalar angle"
        with pytest.raises(TypeError, match=match):
            obj.angle = 1.0


class TestPositiveScalarAngle:
    """
    Tests for the PositiveScalarAngle descriptor class.
    """

    def test_valid(self):
        """
        Test that a positive scalar angle Quantity is accepted.
        """
        obj = MockPositiveScalarAngle()
        obj.angle = 5.0 * u.arcsec
        assert obj.angle == 5.0 * u.arcsec

    def test_negative_error(self):
        """
        Test that a negative angle raises ValueError.
        """
        obj = MockPositiveScalarAngle()
        match = "'angle' must be greater than zero"
        with pytest.raises(ValueError, match=match):
            obj.angle = -1.0 * u.arcsec

    def test_nonscalar_error(self):
        """
        Test that a non-scalar angle array raises ValueError.
        """
        obj = MockPositiveScalarAngle()
        match = "'angle' must be a scalar"
        with pytest.raises(ValueError, match=match):
            # Single-element array passes > 0 check but fails isscalar
            obj.angle = np.array([5.0]) * u.arcsec

    def test_non_angle_units_error(self):
        """
        Test that a Quantity with non-angular units raises ValueError.
        """
        obj = MockPositiveScalarAngle()
        match = "'angle' must have angular units"
        with pytest.raises(ValueError, match=match):
            obj.angle = 5.0 * u.m

    def test_non_quantity_error(self):
        """
        Test that a plain float raises TypeError.
        """
        obj = MockPositiveScalarAngle()
        match = "'angle' must be a scalar angle"
        with pytest.raises(TypeError, match=match):
            obj.angle = 5.0


class TestScalarAngleOrValue:
    """
    Tests for the ScalarAngleOrValue descriptor class.
    """

    def test_valid_quantity(self):
        """
        Test that a scalar angle Quantity is accepted.
        """
        obj = MockScalarAngleOrValue()
        obj.theta = 30.0 * u.deg
        assert obj.theta == 30.0 * u.deg

    def test_float_converted_to_radian(self):
        """
        Test that a plain float is stored as a radian Quantity.
        """
        obj = MockScalarAngleOrValue()
        obj.theta = 1.5
        assert isinstance(obj.theta, u.Quantity)
        assert obj.theta.unit == u.radian
        assert obj.theta.value == 1.5

    def test_reset_lazyproperties(self):
        """
        Test that setting theta twice clears cached lazyproperties.
        """
        obj = MockScalarAngleOrValue()
        obj.theta = 1.0
        obj._lazyproperties = ['cached']
        obj.cached = 42
        obj.theta = 2.0  # triggers reset
        assert 'cached' not in obj.__dict__

    def test_nonscalar_quantity_error(self):
        """
        Test that a non-scalar angle Quantity raises ValueError.
        """
        obj = MockScalarAngleOrValue()
        match = "'theta' must be a scalar"
        with pytest.raises(ValueError, match=match):
            obj.theta = np.array([1.0, 2.0]) * u.arcsec

    def test_non_angle_units_error(self):
        """
        Test that a Quantity with non-angular units raises ValueError.
        """
        obj = MockScalarAngleOrValue()
        match = "'theta' must have angular units"
        with pytest.raises(ValueError, match=match):
            obj.theta = 5.0 * u.m

    def test_nonscalar_float_error(self):
        """
        Test that a non-scalar float array raises ValueError.
        """
        obj = MockScalarAngleOrValue()
        match = 'If not an angle Quantity'
        with pytest.raises(ValueError, match=match):
            obj.theta = np.array([1.0, 2.0])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the model module.
"""

import warnings

import numpy as np
import pytest
from astropy.io import fits
from astropy.modeling.models import Gaussian2D
from astropy.utils.data import get_pkg_data_filename

from photutils.datasets.load import _get_path
from photutils.isophote.ellipse import Ellipse
from photutils.isophote.geometry import EllipseGeometry
from photutils.isophote.isophote import IsophoteList
from photutils.isophote.model import build_ellipse_model
from photutils.isophote.tests.make_test_data import make_test_image


@pytest.mark.remote_data
def test_model():
    path = _get_path('isophote/M105-S001-RGB.fits',
                     location='photutils-datasets', cache=True)
    hdu = fits.open(path)
    data = hdu[0].data[0]
    hdu.close()

    g = EllipseGeometry(530.0, 511, 10.0, 0.1, np.deg2rad(10.0))
    ellipse = Ellipse(data, geometry=g, threshold=1.0e5)

    # NOTE: this sometimes emits warnings (e.g., py38, ubuntu), but
    # sometimes not. Here we simply ignore any RuntimeWarning, whether
    # there is one or not.
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', RuntimeWarning)
        isophote_list = ellipse.fit_image()
    model = build_ellipse_model(data.shape, isophote_list,
                                fill=np.mean(data[10:100, 10:100]))

    assert data.shape == model.shape

    residual = data - model
    assert np.abs(np.mean(residual)) <= 5.0


@pytest.mark.parametrize('sma_interval', [0.05, 0.1])
def test_model_simulated_data(sma_interval):
    data = make_test_image(nx=200, ny=200, i0=10.0, sma=5.0, eps=0.5,
                           pa=np.pi / 3.0, noise=0.05, seed=0)

    g = EllipseGeometry(100.0, 100.0, 5.0, 0.5, np.pi / 3.0)
    ellipse = Ellipse(data, geometry=g, threshold=1.0e5)

    # Catch warnings that may arise from empty slices. This started
    # to happen on windows with scipy 1.15.0.
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', RuntimeWarning)
        isophote_list = ellipse.fit_image()

    model = build_ellipse_model(data.shape, isophote_list,
                                fill=np.mean(data[0:50, 0:50]),
                                sma_interval=sma_interval)

    assert data.shape == model.shape

    residual = data - model
    assert np.abs(np.mean(residual)) <= 0.01
    assert np.abs(np.median(residual)) <= 0.01


def test_model_minimum_radius():
    # This test requires a "defective" image that drives the
    # model building algorithm into a corner, where it fails.
    # With the algorithm fixed, it bypasses the failure and
    # succeeds in building the model image.
    filepath = get_pkg_data_filename('data/minimum_radius_test.fits')
    with fits.open(filepath) as hdu:
        data = hdu[0].data
        g = EllipseGeometry(50.0, 45, 530.0, 0.1, np.deg2rad(10.0))
        g.find_center(data)
        ellipse = Ellipse(data, geometry=g)

        match1 = 'Degrees of freedom'
        match2 = 'Mean of empty slice'
        match3 = 'invalid value encountered'
        ctx1 = pytest.warns(RuntimeWarning, match=match1)
        ctx2 = pytest.warns(RuntimeWarning, match=match2)
        ctx3 = pytest.warns(RuntimeWarning, match=match3)
        with ctx1, ctx2, ctx3:
            isophote_list = ellipse.fit_image(sma0=40, minsma=0,
                                              maxsma=350.0, step=0.4, n_clip=3)

        model = build_ellipse_model(data.shape, isophote_list,
                                    fill=np.mean(data[0:50, 0:50]))

        # It's enough that the algorithm reached this point. The
        # actual accuracy of the modelling is being tested elsewhere.
        assert data.shape == model.shape


def test_model_inputs():
    match = 'isolist must not be empty'
    with pytest.raises(ValueError, match=match):
        build_ellipse_model((10, 10), IsophoteList([]))


def test_model_harmonics():
    """
    Test that high harmonics are included in build_ellipse_model.
    """
    x0 = y0 = 50
    xsig = 10
    ysig = 5
    eps = ysig / xsig
    theta = np.deg2rad(41)
    m = Gaussian2D(100, x0, y0, xsig, ysig, theta)
    yy, xx = np.mgrid[:101, :101]
    data = m(xx, yy)
    yy -= y0
    xx -= x0
    dt = np.arctan2(yy, xx) - np.deg2rad(10)
    harm = (0.1 * np.sin(3 * dt)
            + 0.1 * np.cos(3 * dt)
            + 0.6 * np.sin(4 * dt)
            - 0.5 * np.cos(4 * dt))
    harm -= np.min(harm)
    data += 5 * harm

    geometry = EllipseGeometry(x0=x0, y0=y0, sma=30, eps=eps, pa=theta)
    ellipse = Ellipse(data, geometry=geometry)
    isolist = ellipse.fit_image(fix_center=True, fix_eps=True)

    model_image = build_ellipse_model(data.shape, isolist, high_harmonics=True)
    residual = data - model_image

    mask = model_image > 0
    assert np.std(residual[mask]) < 0.4


def test_model_integration():
    """
    Test that model integration does not stop as soon as the angle reaches
    the edge of the image.
    """
    data = make_test_image(nx=80, ny=110, i0=100.0, sma=60.0, eps=0.5,
                           pa=np.pi / 3.0, noise=0.05, seed=0)
    g = EllipseGeometry(40, 55, 5.0, 0.5, np.pi / 3.0)
    ellipse = Ellipse(data, geometry=g, threshold=1.0e5)
    isophote_list = ellipse.fit_image()
    model = build_ellipse_model(data.shape, isophote_list,
                                fill=np.nanmean(data[105:, :5]),
                                sma_interval=0.05)
    assert np.nanmean(np.abs(model[100:, 60:] - data[100:, 60:])) < 2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the wcs module.
"""

import pytest
from numpy.testing import assert_allclose

from photutils.datasets import make_gwcs, make_wcs
from photutils.utils._optional_deps import HAS_GWCS


def test_make_wcs():
    shape = (100, 200)
    wcs = make_wcs(shape)

    assert wcs.pixel_shape == shape
    assert wcs.wcs.radesys == 'ICRS'

    wcs = make_wcs(shape, galactic=True)
    assert wcs.wcs.ctype[0] == 'GLON-CAR'
    assert wcs.wcs.ctype[1] == 'GLAT-CAR'


@pytest.mark.skipif(not HAS_GWCS, reason='gwcs is required')
def test_make_gwcs():
    shape = (100, 200)

    wcs = make_gwcs(shape)
    assert wcs.pixel_n_dim == 2
    assert wcs.available_frames == ['detector', 'icrs']
    assert wcs.output_frame.name == 'icrs'
    assert wcs.output_frame.axes_names == ('lon', 'lat')

    wcs = make_gwcs(shape, galactic=True)
    assert wcs.pixel_n_dim == 2
    assert wcs.available_frames == ['detector', 'galactic']
    assert wcs.output_frame.name == 'galactic'
    assert wcs.output_frame.axes_names == ('lon', 'lat')


@pytest.mark.skipif(not HAS_GWCS, reason='gwcs is required')
def test_make_wcs_compare():
    shape = (200, 300)
    wcs = make_wcs(shape)
    gwcs_obj = make_gwcs(shape)
    sc1 = wcs.pixel_to_world((50, 75), (50, 100))
    sc2 = gwcs_obj.pixel_to_world((50, 75), (50, 100))

    assert_allclose(sc1.ra, sc2.ra)
    assert_allclose(sc1.dec, sc2.dec)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Pytest configuration and shared fixtures for psf_matching tests.
"""

import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D


def _make_gaussian_psf(size, std):
    """
    Make a centered, normalized 2D Gaussian PSF.
    """
    cen = (size - 1) / 2.0
    yy, xx = np.mgrid[0:size, 0:size]
    model = Gaussian2D(1.0, cen, cen, std, std)
    psf = model(xx, yy)
    return psf / psf.sum()


def _make_gaussian_psf_noncentered(shape, std, xcen, ycen):
    """
    Make a non-centered, normalized 2D Gaussian PSF.
    """
    yy, xx = np.mgrid[0:shape[0], 0:shape[1]]
    model = Gaussian2D(1.0, xcen, ycen, std, std)
    psf = model(xx, yy)
    return psf / psf.sum()


@pytest.fixture(name='psf1')
def psf1():
    """
    Narrow Gaussian PSF (source).
    """
    return _make_gaussian_psf(25, 3.0)


@pytest.fixture(name='psf2')
def psf2():
    """
    Broad Gaussian PSF (target).
    """
    return _make_gaussian_psf(25, 5.0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for plotting Gridded PSF models.
"""

import numpy as np
from astropy.visualization import simple_norm

__all__ = []


def _plot_grid_docstring(func):
    func.__doc__ = """
        Plot the grid of ePSF models.

        Parameters
        ----------
        ax : `matplotlib.axes.Axes` or `None`, optional
            The matplotlib axes on which to plot. If `None`, then the
            current `~matplotlib.axes.Axes` instance is used.

        vmax_scale : float, optional
            Scale factor to apply to the image stretch limits. This
            value is multiplied by the peak ePSF value to determine the
            plotting ``vmax``. The defaults are 1.0 for plotting the
            ePSF data and 0.03 for plotting the ePSF difference data
            (``deltas=True``). If ``deltas=True``, the ``vmin`` is set
            to ``-vmax``. If ``deltas=False`` the ``vmin`` is set to
            ``vmax`` / 1e4.

        peak_norm : bool, optional
            Whether to normalize the ePSF data by the peak value. The
            default shows the ePSF flux per pixel.

        deltas : bool, optional
            Set to `True` to show the differences between each ePSF
            and the average ePSF.

        cmap : str or `matplotlib.colors.Colormap`, optional
            The colormap to use. The default is 'viridis'.

        dividers : bool, optional
            Whether to show divider lines between the ePSFs.

        divider_color, divider_ls : str, optional
            Matplotlib color and linestyle options for the divider
            lines between ePSFs. These keywords have no effect unless
            ``show_dividers=True``.

        figsize : (float, float), optional
            The figure (width, height) in inches.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            The matplotlib figure object. This will be the current
            figure if ``ax=None``. Use ``fig.savefig()`` to save the
            figure to a file.

        Notes
        -----
        This method returns a figure object. If you are using this
        method in a script, you will need to call ``fig.show()`` to
        display the figure. If you are using this method in a Jupyter
        notebook, the figure will be displayed automatically.

        When in a notebook, if you do not store the return value of this
        function, the figure will be displayed twice due to the REPL
        automatically displaying the return value of the last function
        call. Alternatively, you can append a semicolon to the end of
        the function call to suppress the display of the return value.
        """
    return func


class _ModelGridPlotter:
    """
    Class to plot a grid of ePSF models.
    """

    def __init__(self, model):
        self.model = model

    def _reshape_grid(self, data):
        """
        Reshape the 3D ePSF grid as a 2D array of horizontally and
        vertically stacked ePSFs.

        Parameters
        ----------
        data : `numpy.ndarray`
            The 3D array of ePSF data.

        Returns
        -------
        reshaped_data : `numpy.ndarray`
            The 2D array of ePSF data.
        """
        nypsfs = self.model._ygrid.shape[0]
        nxpsfs = self.model._xgrid.shape[0]
        ny, nx = self.model.data.shape[1:]
        return (data.reshape(nypsfs, nxpsfs, ny, nx)
                .transpose([0, 2, 1, 3])
                .reshape(nypsfs * ny, nxpsfs * nx))

    @_plot_grid_docstring
    def plot_grid(self, *, ax=None, vmax_scale=None, peak_norm=False,
                  deltas=False, cmap='viridis', dividers=True,
                  divider_color='darkgray', divider_ls='-', figsize=None):
        import matplotlib.pyplot as plt
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        data = self.model.data.copy()
        if deltas:
            # Compute mean ignoring any blank (all zeros) ePSFs.
            # This is the case for MIRI with its non-square FOV.
            mask = np.zeros(data.shape[0], dtype=bool)
            for i, arr in enumerate(data):
                if np.count_nonzero(arr) == 0:
                    mask[i] = True
            data -= np.mean(data[~mask], axis=0)
            data[mask] = 0.0

        data = self._reshape_grid(data)

        if ax is None:
            if (figsize is None
                    and self.model.meta.get('detector', '') == 'NRCSW'):
                figsize = (20, 8)
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = ax.get_figure()

        if peak_norm and data.max() != 0:
            # normalize relative to peak
            data /= data.max()

        if deltas:
            if vmax_scale is None:
                vmax_scale = 0.03
            vmax = data.max() * vmax_scale
            vmin = -vmax
            norm = simple_norm(data, 'linear', vmin=vmin, vmax=vmax)
        else:
            if vmax_scale is None:
                vmax_scale = 1.0
            vmax = data.max() * vmax_scale
            vmin = vmax / 1.0e4
            norm = simple_norm(data, 'log', vmin=vmin, vmax=vmax, log_a=1.0e4)

        # Set up the coordinate axes to later set tick labels based on
        # detector ePSF coordinates. This sets up axes to have, behind the
        # scenes, the ePSFs centered at integer coords 0, 1, 2, 3 etc.
        # extent order: left, right, bottom, top
        nypsfs = self.model._ygrid.shape[0]
        nxpsfs = self.model._xgrid.shape[0]
        extent = [-0.5, nxpsfs - 0.5, -0.5, nypsfs - 0.5]

        axim = ax.imshow(data, extent=extent, norm=norm, cmap=cmap,
                         origin='lower')

        # Use the axes set up above to set appropriate tick labels
        xticklabels = self.model._xgrid.astype(int)
        yticklabels = self.model._ygrid.astype(int)
        if self.model.meta.get('detector', '') == 'NRCSW':
            xticklabels = list(xticklabels[0:5]) * 4
            yticklabels = list(yticklabels[0:5]) * 2
        ax.set_xticks(np.arange(nxpsfs))
        ax.set_xticklabels(xticklabels)
        ax.set_xlabel('ePSF location in detector X pixels')
        ax.set_yticks(np.arange(nypsfs))
        ax.set_yticklabels(yticklabels)
        ax.set_ylabel('ePSF location in detector Y pixels')

        if dividers:
            for ix in range(nxpsfs - 1):
                ax.axvline(ix + 0.5, color=divider_color, ls=divider_ls)
            for iy in range(nypsfs - 1):
                ax.axhline(iy + 0.5, color=divider_color, ls=divider_ls)

        instrument = self.model.meta.get('instrument', '')
        if not instrument:
            # WebbPSF output
            instrument = self.model.meta.get('instrume', '')
        detector = self.model.meta.get('detector', '')
        filtername = self.model.meta.get('filter', '')

        # WebbPSF outputs a tuple with the comment in the second element
        if isinstance(instrument, (tuple, list, np.ndarray)):
            instrument = instrument[0]
        if isinstance(detector, (tuple, list, np.ndarray)):
            detector = detector[0]
        if isinstance(filtername, (tuple, list, np.ndarray)):
            filtername = filtername[0]

        title = f'{instrument} {detector} {filtername}'
        if title != '':
            # add extra space at end
            title += ' '

        if deltas:
            minus = '\u2212'
            ax.set_title(f'{title}(ePSFs {minus} <ePSF>)')
            if peak_norm:
                label = 'Difference relative to average ePSF peak'
            else:
                label = 'Difference relative to average ePSF values'
        else:
            ax.set_title(f'{title}ePSFs')
            if peak_norm:
                label = 'Scale relative to ePSF peak pixel'
            else:
                label = 'ePSF flux per pixel'

        divider = make_axes_locatable(ax)
        cax_cbar = divider.append_axes('right', size='3%', pad='3%')
        cbar = fig.colorbar(axim, cax=cax_cbar, label=label)

        if not deltas:
            cbar.ax.set_yscale('log')

        if self.model.meta.get('detector', '') == 'NRCSW':
            # NIRCam NRCSW STDPSF files contain all detectors.
            # The plot gets extra divider lines and SCA name labels.
            nxpsfs = len(self.model._xgrid)
            nypsfs = len(self.model._ygrid)
            ax.axhline(nypsfs / 2 - 0.5, color='orange')
            for i in range(1, 4):
                ax.axvline(nxpsfs / 4 * i - 0.5, color='orange')

            det_labels = [['A1', 'A3', 'B4', 'B2'], ['A2', 'A4', 'B3', 'B1']]
            for i in range(2):
                for j in range(4):
                    ax.text(j * nxpsfs / 4 - 0.45,
                            (i + 1) * nypsfs / 2 - 0.55,
                            det_labels[i][j], color='orange',
                            verticalalignment='top', fontsize=12)

        fig.tight_layout()

        return fig



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for image segmentation.
"""

import numpy as np
from astropy.convolution import Gaussian2DKernel
from astropy.stats import gaussian_fwhm_to_sigma
from scipy.ndimage import generate_binary_structure

from photutils.utils._deprecation import deprecated_positional_kwargs
from photutils.utils._parameters import as_pair

__all__ = ['make_2dgaussian_kernel']


@deprecated_positional_kwargs(since='3.0', until='4.0')
def make_2dgaussian_kernel(fwhm, size, mode='oversample', oversampling=10):
    """
    Make a normalized 2D circular Gaussian kernel.

    The kernel must have odd sizes in both X and Y, be centered in the
    central pixel, and normalized to sum to 1.

    Parameters
    ----------
    fwhm : float
        The full-width at half-maximum (FWHM) of the 2D circular
        Gaussian kernel.

    size : int or (2,) int array_like
        The size of the kernel along each axis. If ``size`` is a scalar
        then a square size of ``size`` will be used. If ``size`` has
        two elements, they must be in ``(ny, nx)`` (i.e., array shape)
        order. ``size`` must have odd values for both axes.

    mode : {'oversample', 'center', 'linear_interp', 'integrate'}, optional
        The mode to use for discretizing the 2D Gaussian model:

        * 'oversample' (default):
          Discretize model by taking the average on an oversampled grid.

        * 'center':
          Discretize model by taking the value at the center of the bin.

        * 'linear_interp':
          Discretize model by performing a bilinear interpolation
          between the values at the corners of the bin.

        * 'integrate':
          Discretize model by integrating the model over the bin.

    oversampling : int, optional
        The oversampling factor used when ``mode='oversample'``.

    Returns
    -------
    kernel : `astropy.convolution.Kernel2D`
        The output smoothing kernel, normalized such that it sums to 1.
    """
    ysize, xsize = as_pair('size', size, lower_bound=(0, 0), check_odd=True)

    kernel = Gaussian2DKernel(fwhm * gaussian_fwhm_to_sigma,
                              x_size=xsize, y_size=ysize, mode=mode,
                              factor=oversampling)
    kernel.normalize(mode='integral')  # ensure kernel sums to 1

    return kernel


def _make_binary_structure(ndim, connectivity):
    """
    Make a binary structure element.

    Parameters
    ----------
    ndim : int
        The number of array dimensions.

    connectivity : {4, 8}
        For the case of ``ndim=2``, the type of pixel connectivity used
        in determining how pixels are grouped into a detected source.
        The options are 4 or 8 (default). 4-connected pixels touch along
        their edges. 8-connected pixels touch along their edges or
        corners. For reference, SourceExtractor uses 8-connected pixels.

    Returns
    -------
    array : `~numpy.ndarray`
        The binary structure element. If ``ndim <= 2`` an array of int
        is returned, otherwise an array of bool is returned.
    """
    if ndim == 1:
        footprint = np.array((1, 1, 1))
    elif ndim == 2:
        if connectivity == 4:
            footprint = np.array(((0, 1, 0), (1, 1, 1), (0, 1, 0)))
        elif connectivity == 8:
            footprint = np.ones((3, 3), dtype=int)
        else:
            msg = f'Invalid connectivity={connectivity} -- options are 4 or 8'
            raise ValueError(msg)
    else:
        footprint = generate_binary_structure(ndim, 1)

    return footprint


def _mask_to_mirrored_value(data, replace_mask, xycenter, *, mask=None):
    """
    Replace masked pixels with the value of the pixel mirrored across a
    given center position.

    If the mirror pixel is unavailable (i.e., it is outside the image or
    masked), then the masked pixel value is set to zero.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        A 2D array.

    replace_mask : 2D bool `~numpy.ndarray`
        A boolean mask where `True` values indicate the pixels that
        should be replaced, if possible, by mirrored pixel values. It
        must have the same shape as ``data``.

    xycenter : tuple of two int
        The (x, y) center coordinates around which masked pixels will be
        mirrored.

    mask : 2D bool `~numpy.ndarray`
        A boolean mask where `True` values indicate ``replace_mask``
        *mirrored* pixels that should never be used to fix
        ``replace_mask`` pixels. In other words, if a pixel in
        ``replace_mask`` has a mirror pixel in this ``mask``, then the
        mirrored value is set to zero. Using this keyword prevents
        potential spreading of known non-finite or bad pixel values.

    Returns
    -------
    result : 2D `~numpy.ndarray`
        A 2D array with replaced masked pixels.
    """
    outdata = np.copy(data)

    ymasked, xmasked = np.nonzero(replace_mask)
    xmirror = 2 * int(xycenter[0] + 0.5) - xmasked
    ymirror = 2 * int(xycenter[1] + 0.5) - ymasked

    # Find mirrored pixels that are outside the image
    badmask = ((xmirror < 0) | (ymirror < 0) | (xmirror >= data.shape[1])
               | (ymirror >= data.shape[0]))

    # Remove them from the set of replace_mask pixels and set them to
    # zero
    if np.any(badmask):
        outdata[ymasked[badmask], xmasked[badmask]] = 0.0
        # Remove the badmask pixels from pixels to be replaced
        goodmask = ~badmask
        ymasked = ymasked[goodmask]
        xmasked = xmasked[goodmask]
        xmirror = xmirror[goodmask]
        ymirror = ymirror[goodmask]

    outdata[ymasked, xmasked] = outdata[ymirror, xmirror]

    # Find mirrored pixels that are masked and replace_mask pixels that are
    # mirrored to other replace_mask pixels. Set them both to zero.
    mirror_mask = replace_mask[ymirror, xmirror]
    if mask is not None:
        mirror_mask |= mask[ymirror, xmirror]
    xbad = xmasked[mirror_mask]
    ybad = ymasked[mirror_mask]
    outdata[ybad, xbad] = 0.0

    return outdata
